import re
import inspect

from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, InvocationJob
from lifeblood.invocationjob import InvocationRequirements
from lifeblood.enums import NodeParameterType, WorkerType
from lifeblood.text import filter_by_pattern

from typing import Iterable


def node_class():
    return HoudiniDistributedTracker


async def do(port, webport, killport):
    import asyncio
    import os
    import signal
    import socket
    import lifeblood_connection
    import json

    close_event = asyncio.Event()

    async def conn_cb(reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        await reader.readexactly(1)
        print('got termination request')
        writer.close()
        close_event.set()

    ip = lifeblood_connection.get_host_ip()
    # find free ports
    port = lifeblood_connection.get_free_tcp_port(ip, port)
    webport = lifeblood_connection.get_free_tcp_port(ip, webport)
    if webport == port:
        webport = lifeblood_connection.get_free_tcp_port(ip, webport + 1)
    killport = lifeblood_connection.get_free_tcp_port(ip, killport)
    if killport == port:
        killport = lifeblood_connection.get_free_tcp_port(ip, killport + 1)
    if killport == webport:  # NOTE: port < webport, always, so we need 2 independent checks. worst case both conditions may be true
        killport = lifeblood_connection.get_free_tcp_port(ip, killport + 1)

    # at this point we have free ports, but by the time we start our servers - ports might get taken

    server = await asyncio.start_server(conn_cb, port=killport, family=socket.AF_INET)
    proc = await asyncio.create_subprocess_exec('hython', '-m', 'simtracker', '-v', str(port), str(webport))

    async def server_close_waiter(server: asyncio.AbstractServer):
        await close_event.wait()
        server.close()
        await server.wait_closed()

    async def proc_end_waiter(proc: asyncio.subprocess.Process):
        return await proc.wait()

    await server.start_serving()

    attrs = json.loads(os.environ['LBATTRS_JSON'])
    attrs['simtracker_host'] = ip
    attrs['simtracker_port'] = port
    attrs['tracker kill port'] = killport
    lifeblood_connection.create_task('spawned task', attrs)
    await asyncio.wait([proc_end_waiter(proc), server_close_waiter(server)], return_when=asyncio.FIRST_COMPLETED)
    print(f'done, exiting!')

    proc_error = proc.returncode  # if simtracker process had error - return it, otherwise we terminate it and don't care about it's return code
    if proc_error is None:
        proc.terminate()
        await proc.wait()
        print('wait done')
    print('closed simtracker')
    if server.is_serving():
        server.close()
        await server.wait_closed()
    print('closed server')

    return proc_error or 0  # read above - None counts as good as we terminate it ourselves


class HoudiniDistributedTracker(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'houdini sim tracker'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'houdini', 'tracker', 'distributed'

    @classmethod
    def type_name(cls) -> str:
        return 'houdini_distributed_tracker'

    @classmethod
    def description(cls) -> str:
        return 'creates a running houdini distributed simulation tracker\n' \
               'when tracker is up and running - a child task is created\n' \
               'child task inherits all attributes from parent\n' \
               'additionally these attribs are created:\n' \
               '    SIMTRACKER_HOST: hostname of the tracker\n' \
               '    SIMTRACKER_PORT: port of the tracker\n' \
               '    tracker kill port: port of the tracker for control commands'

    def __init__(self, name: str):
        super(HoudiniDistributedTracker, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_output('spawned')
            ui.add_parameter('port', 'sim port', NodeParameterType.INT, 19375)
            ui.add_parameter('wport', 'web port', NodeParameterType.INT, 19376)
            ui.add_parameter('kport', 'kill port', NodeParameterType.INT, 19377)
            ui.add_parameter('use helper', 'run on scheduler helper', NodeParameterType.BOOL, True)

    def process_task(self, context) -> ProcessingResult:
        code = 'import sys\n' \
               'import asyncio\n\n'
        code += inspect.getsource(do)
        code += '\n' \
                'port, webport, killport = [int(x) for x in sys.argv[1:4]]\n' \
                'sys.exit(asyncio.get_event_loop().run_until_complete(do(port, webport, killport)))\n'
        invoc = InvocationJob(['python', ':/work_to_do.py', context.param_value('port'), context.param_value('wport'), context.param_value('kport')])
        invoc.set_extra_file('work_to_do.py', code)

        if context.param_value('use helper'):
            invoc.set_requirements(InvocationRequirements(worker_type=WorkerType.SCHEDULER_HELPER))
        res = ProcessingResult(invoc)

        return res



import time

from lifeblood.basenode import BaseNode
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.nodethings import ProcessingResult
from lifeblood.uidata import NodeParameterType

from typing import TYPE_CHECKING, Iterable
if TYPE_CHECKING:
    from lifeblood.scheduler import Scheduler


def node_class():
    return Test


class Test(BaseNode):
    def __init__(self, name):
        super(Test, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('multiline text', 'wow multi line', NodeParameterType.STRING, 'bla').set_text_multiline()
            ui.add_parameter('multiline python text', 'wow multi line code', NodeParameterType.STRING, '# shize').set_text_multiline('python')
            ui.add_parameter('test readonly crap', 'lelabel', NodeParameterType.STRING, 'wow im text', readonly=True)
            with ui.parameters_on_same_line_block():
                ui.add_parameter('text1', 'lobel', NodeParameterType.STRING, 'hi text i\'m dad', readonly=True)
                ui.add_parameter('text2', 'lebel', NodeParameterType.FLOAT, 4.22, readonly=True)
                ui.add_parameter('text3', 'lubel', NodeParameterType.INT, 1, readonly=True).add_menu([('fee', 0), ('foo', 1), ('faa', 2)])
                ui.add_parameter('text4', 'ha, and i\'m editable', NodeParameterType.STRING, 'yes')
            ui.add_parameter('some int or smth', 'into', NodeParameterType.INT, -1)
            ui.add_parameter('param name', 'label for display', NodeParameterType.STRING, 'default value')
            ui.add_parameter('fooof', 'param with menu', NodeParameterType.INT, 0).add_menu((('display val1', 0), ('display val2', 4), ('donkey!', -2)))

            with ui.parameters_on_same_line_block():
                ui.add_parameter('name uniqye', 'Name', NodeParameterType.STRING, '')
                ui.add_parameter('typeasss', '', NodeParameterType.INT, 0)

            with ui.multigroup_parameter_block('attr_count'):
                ui.add_parameter('smth41', 'label', NodeParameterType.INT, 0)

    @classmethod
    def label(cls) -> str:
        return 'test node'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'test', 'debug'

    @classmethod
    def type_name(cls) -> str:
        return 'test'

    def process_task(self, context) -> ProcessingResult:
        td = InvocationJob(['bash', '-c',
                            'echo "startin..."\n'
                            'for i in {1..15}\n'
                            'do\n'
                            '    echo "iteration $i"\n'
                            '    echo $(date)\n'
                            '    sleep 1\n'
                            'done\n'
                            'echo "ended"\n'],
                           env=InvocationEnvironment())
        time.sleep(6)  # IMITATE LAUNCHING LONG BLOCKING OPERATION
        if int(time.time()) % 10 == 0:
            raise RuntimeError('just happened to be in a wrong time in a wrong place')
        return ProcessingResult(job=td)

    def postprocess_task(self, context) -> ProcessingResult:
        time.sleep(3.5)  # IMITATE LAUNCHING LONG BLOCKING OPERATION
        res = ProcessingResult()
        res.set_attribute('cat', 1)
        res.set_attribute('dog', 2)
        if int(time.time()) % 10 == 0:
            raise RuntimeError('ain\'t ur lucky day, kid')
        return res



import asyncio
from .logging import get_logger
#from .scheduler_task_protocol import SchedulerTaskClient
from .scheduler_message_processor import SchedulerWorkerControlClient
from .net_messages.address import AddressChain
from .net_messages.message_processor import MessageProcessorBase
from .net_messages.exceptions import MessageTransferError

from typing import Tuple, Callable, Coroutine


class PulseChecker:
    def __init__(self, address: AddressChain, message_processor: MessageProcessorBase, interval: float = 5, maximum_misses: int = 5):
        self.__address = address
        self.__message_processor = message_processor
        self.__interval = interval
        self.__pinger_task = None
        self.__stop_event = asyncio.Event()
        self.__misses = 0
        self.__maximum_misses = maximum_misses
        self.__miss_reported = False
        self.__logger = get_logger('Pulse')

        self.__on_fail_callbacks = set()

    async def start(self):
        self.__pinger_task = asyncio.create_task(self.pinger())
        # TODO: handle initial failing to start

    def stop(self):
        if self.__pinger_task is None:
            raise RuntimeError('not started')
        self.__stop_event.set()

    async def wait_till_stops(self):
        if self.__pinger_task is None:
            raise RuntimeError('not started')
        await self.__pinger_task

    def __await__(self):
        return self.wait_till_stops().__await__()

    def add_pulse_fail_callback(self, async_func: Callable[[], Coroutine]):
        self.__on_fail_callbacks.add(async_func)

    def remove_pulse_fail_callback(self, async_func: Callable[[], Coroutine]):
        self.__on_fail_callbacks.remove(async_func)

    async def pinger(self):
        stop_waiter = asyncio.create_task(self.__stop_event.wait())
        while not self.__stop_event.is_set():
            done, _ = await asyncio.wait([stop_waiter], timeout=self.__interval)
            if stop_waiter in done:
                break

            try:
                with SchedulerWorkerControlClient.get_scheduler_control_client(self.__address, self.__message_processor) as client:  # type: SchedulerWorkerControlClient
                    await client.pulse()
                    self.__misses = 0
                    if self.__miss_reported:
                        self.__logger.info('pulse restored')
                        self.__miss_reported = False
            except MessageTransferError:
                self.__misses += 1
                self.__logger.warning(f'scheduler missed pulse, current miss count: {self.__misses}')

            if self.__misses >= self.__maximum_misses:
                self.__logger.warning(f'shceduler missed {self.__misses} pulses, reporting')
                self.__miss_reported = True
                for func in self.__on_fail_callbacks:
                    await func()



import os
import asyncio
import aiofiles
from contextlib import contextmanager
from .exceptions import NotEnoughResources, ProcessInitializationError, WorkerNotAvailable, \
    InvocationMessageWrongInvocationId, InvocationMessageAddresseeTimeout, InvocationMessageError
from .environment_resolver import ResolutionImpossibleError
from . import logging
from . import invocationjob
from .exceptions import AlreadyRunning
from .enums import WorkerPingReply, TaskScheduleStatus, InvocationMessageResult
from .net_messages.impl.tcp_simple_command_message_processor import TcpCommandMessageProcessor
from .net_messages.impl.clients import CommandJsonMessageClient
from .net_messages.exceptions import MessageTransferTimeoutError, MessageTransferError
from .net_messages.address import AddressChain
from .net_messages.messages import Message
from .net_messages.impl.message_haldlers import CommandMessageHandlerBase


from typing import Optional, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from .worker import Worker


class WorkerCommandHandler(CommandMessageHandlerBase):
    def __init__(self, worker: "Worker"):
        super().__init__()
        self.__logger = logging.get_logger('worker.message_handler')
        self.__worker = worker

    def command_mapping(self):
        return {'ping': self._command_ping,
                'task': self._command_task,
                'quit': self._command_quit,
                'drop': self._command_drop,
                'status': self._command_status,
                'log': self._command_log,
                'invocation_message': self._command_invocation_message}

    #
    # commands
    #

    #
    # command ping
    async def _command_ping(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
        returns keys:
            ps: ping status
            pv: task completion percentage (0-1) if any
        """
        # TODO: implement this shit too
        # if self.__worker.is_stopping():
        #     pstats = WorkerPingReply.OFF.value
        #     pvalue = 0
        # el
        if self.__worker.is_task_running():
            pstatus = WorkerPingReply.BUSY.value
            pvalue = int(self.__worker.task_status() or 0)
        else:
            pstatus = WorkerPingReply.IDLE.value
            pvalue = 0
        await client.send_message_as_json({'ps': pstatus,
                                           'pv': pvalue})

    #
    # command enqueue task
    async def _command_task(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
            task: serialized InvocationJob
            reply_to: AddressChain where to reply about completion
        returns keys:
            status: TaskScheduleStatus, status of the operation
        """
        task = invocationjob.InvocationJob.deserialize(args['task'].encode('latin1'))
        addr = AddressChain(args['reply_to']) if args.get('reply_to') else original_message.message_source()
        reply = {}

        self.__logger.debug(f'got task: {task}, reply result to {addr}')
        try:
            self.__logger.debug('taking the task')
            await self.__worker.run_task(task, addr)
            reply['status'] = TaskScheduleStatus.SUCCESS.value
        except AlreadyRunning:
            self.__logger.debug('BUSY. rejecting task')
            reply['status'] = TaskScheduleStatus.BUSY.value
        except ResolutionImpossibleError:
            self.__logger.info('Worker failed to resolve required environment. rejecting task')
            reply['status'] = TaskScheduleStatus.FAILED.value
        except ProcessInitializationError:
            self.__logger.info('Failed to initialize payload process. rejecting task')
            reply['status'] = TaskScheduleStatus.FAILED.value
        except NotEnoughResources:
            self.__logger.warning('Not enough resources (this is unusual error - scheduler should know our resources). rejecting task')
            reply['status'] = TaskScheduleStatus.FAILED.value
        except WorkerNotAvailable:
            self.__logger.warning('Got a task, but Worker is not available. Most probably is stopping right now')
            reply['status'] = TaskScheduleStatus.FAILED.value
        except Exception as e:
            self.__logger.exception('no, cuz %s', e)
            reply['status'] = TaskScheduleStatus.FAILED.value

        await client.send_message_as_json(reply)

    #
    # quit worker
    async def _command_quit(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
        returns keys:
        """
        self.__worker.stop()
        await client.send_message_as_json({})

    #
    # command drop/cancel current task
    async def _command_drop(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
        returns keys:
        """
        try:
            await self.__worker.cancel_task()
        except Exception:
            self.__logger.exception('task drop failed')
        await client.send_message_as_json({})

    #
    # command check worker status
    async def _command_status(self, reader, args: dict, client: CommandJsonMessageClient, original_message: Message):
        raise NotImplementedError()

    #
    # command to get worker's logs
    async def _command_log(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
            invoc_id: invocation id to get logs of
        returns keys:
            stdout:
            stderr:
        """
        invocation_id = args['invoc_id']
        result = {}
        for key, logfilepath in (('stdout', self.__worker.get_log_filepath('output', invocation_id)),
                                 ('stderr', self.__worker.get_log_filepath('error', invocation_id))):
            if not os.path.exists(logfilepath):
                result[key] = ''
                continue
            async with aiofiles.open(logfilepath, 'r') as f:
                all_data = await f.read()  #TODO: what if there is binary crap somehow in the log?
                result[key] = all_data

        await client.send_message_as_json(result)

    #
    # commands for inter-task communication
    async def _command_invocation_message(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
            dst_invoc_id: receiver's invocation id
            src_invoc_id: sender's invocation id
            addressee: address id, where to address message within worker
            message_data_raw: message_data_raw
            addressee_timeout: timeout in seconds of how long to wait for addressee to start receiving
        returns keys:
            result: str, operation result
        """
        result = 'unknown'
        try:
            await self.__worker.deliver_invocation_message(args['dst_invoc_id'],
                                                           args['addressee'],
                                                           args['src_invoc_id'],
                                                           args['message_data_raw'].encode('latin1'),
                                                           args['addressee_timeout'])
            result = InvocationMessageResult.DELIVERED.value
        except InvocationMessageWrongInvocationId:
            # it is possible that between sched checking for inv id and message received by worker
            # invocation finished, and we have to catch and report it
            # we report that iid is not running anymore
            result = InvocationMessageResult.ERROR_IID_NOT_RUNNING.value
        except InvocationMessageAddresseeTimeout:
            # we waited enough for addressee to start listening
            result = InvocationMessageResult.ERROR_RECEIVER_TIMEOUT.value
        except Exception:
            result = InvocationMessageResult.ERROR_UNEXPECTED.value

        await client.send_message_as_json({
            'result': result
        })


class WorkerMessageProcessor(TcpCommandMessageProcessor):
    def __init__(self, worker: "Worker", listening_address: Tuple[str, int], *, backlog=4096, connection_pool_cache_time=300):
        super().__init__(listening_address,
                         backlog=backlog,
                         connection_pool_cache_time=connection_pool_cache_time,
                         message_handlers=(WorkerCommandHandler(worker),))


#
# Client
#


class WorkerControlClient:
    def __init__(self, client: CommandJsonMessageClient):
        self.__client = client

    @classmethod
    @contextmanager
    def get_worker_control_client(cls, worker_address: AddressChain, processor: TcpCommandMessageProcessor) -> "WorkerControlClient":
        with processor.message_client(worker_address) as message_client:
            yield WorkerControlClient(message_client)

    async def ping(self) -> Tuple[WorkerPingReply, float]:
        await self.__client.send_command('ping', {})

        reply_message = await self.__client.receive_message()
        data_json = await reply_message.message_body_as_json()
        return WorkerPingReply(data_json['ps']), float(data_json['pv'])

    async def give_task(self, task: invocationjob.InvocationJob, reply_address: Optional[AddressChain] = None) -> TaskScheduleStatus:
        """
        if reply_address is not given - message source address will be used
        """
        await self.__client.send_command('task', {
            'task': (await task.serialize_async()).decode('latin1'),
            'reply_to': str(reply_address) if reply_address else None
        })

        reply = await (await self.__client.receive_message()).message_body_as_json()
        return TaskScheduleStatus(reply['status'])

    async def quit_worker(self):
        await self.__client.send_command('quit', {})

        await self.__client.receive_message()

    async def cancel_task(self) -> None:
        await self.__client.send_command('drop', {})

        await self.__client.receive_message()

    async def status(self):
        raise NotImplementedError()

    async def get_log(self, invocation_id) -> Tuple[str, str]:
        await self.__client.send_command('log', {
            'invoc_id': invocation_id
        })

        reply = await (await self.__client.receive_message()).message_body_as_json()
        return str(reply['stdout']), str(reply['stderr'])

    async def send_invocation_message(self,
                                      destination_invocation_id: int,
                                      destination_addressee: str,
                                      source_invocation_id: Optional[int],
                                      message_body: bytes,
                                      addressee_timeout: float,
                                      overall_timeout: float) -> InvocationMessageResult:
        """
        Note that this command, unlike others, does not raise,
        instead it wraps errors into InvocationMessageResult
        """
        await self.__client.send_command('invocation_message', {
            'dst_invoc_id': destination_invocation_id,
            'src_invoc_id': source_invocation_id,
            'addressee': destination_addressee,
            'addressee_timeout': addressee_timeout,
            'message_data_raw': message_body.decode('latin1'),
        })

        reply = await (await self.__client.receive_message(timeout=overall_timeout)).message_body_as_json()
        return InvocationMessageResult(reply['result'])



import asyncio
import dataclasses

import aiosqlite
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from dataclasses import dataclass
import time
from enum import Enum
from queue import Queue
from ..logging import get_logger
from ..misc import atimeit, aperformance_measurer
from ..enums import InvocationState, TaskState, TaskGroupArchivedState, WorkerState, WorkerType, UIEventType
from ..exceptions import NotSubscribedError
from ..scheduler_event_log import SchedulerEventLog
from ..ui_events import TaskEvent, TaskFullState, TasksUpdated, TasksRemoved, TasksChanged
from ..ui_protocol_data import TaskBatchData, UiData, TaskGroupData, TaskGroupBatchData, TaskGroupStatisticsData, \
    NodeGraphStructureData, WorkerBatchData, WorkerData, WorkerResources, NodeConnectionData, NodeData, TaskData, TaskDelta
from .scheduler_component_base import SchedulerComponentBase
from .data_access import DataAccess

from typing import Dict, Iterable, List, Optional, Tuple, TYPE_CHECKING, Set, Union

if TYPE_CHECKING:  # TODO: maybe separate a subset of scheduler's methods to smth like SchedulerData class, or idunno, for now no obvious way to separate, so having a reference back
    from .scheduler import Scheduler


class QueueEventType(Enum):
    ADDED = 0
    UPDATED = 1
    REMOVED = 2


@dataclass
class LogSubscription:
    expiration_timestamp: float
    event_log: SchedulerEventLog

    def is_expired(self) -> bool:
        return time.time() >= self.expiration_timestamp


class UIStateAccessor(SchedulerComponentBase):
    def __init__(self, scheduler: "Scheduler"):
        super().__init__(scheduler)
        self.__logger = get_logger('scheduler.ui_state_accessor')
        self.__data_access = scheduler.data_access
        self.__ui_cache = {'groups': {}, 'last_update_time': None}

        self.__global_next_event_id = 0  # "atomic" counter of events

        # config
        self.__housekeeping_interval = 60

        # for ui
        self.__latest_graph_ui_state: Optional[NodeGraphStructureData] = None
        self.__latest_graph_ui_event_id = 0
        #
        # the logic here: ui may request updates for (g1, g2, skip_dead) - so that's our key for the log,
        # full updates go there.
        # but individual task events may go to multiple logs, e.g. to (g1, g2), (g1,), (g0, g1) ...
        self.__task_group_event_logs: Dict[Tuple[Tuple[str, ...], bool], LogSubscription] = {}  # maps keys as requested by viewer to logs
        self.__task_group_to_logs: Dict[str, List[LogSubscription]] = {}  # maps task group to list of corresponding logs
        self.__pruning_tasks = []  # TODO: these tasks need to be cancelled clearly on stop

        self.__task_event_preprocess_queue = asyncio.Queue()
        #
        self.__task_group_mapping: Optional[Dict[int, Set[str]]] = None
        self.__group_task_mapping: Optional[Dict[str, Set[id]]] = None
        self.__task_group_mapping_update_alock = asyncio.Lock()
        self.__requested_task_group_force_refresh = True

        # a pool of pools to execute log commands in. the point is to ensure log always runs on ONE SAME THREAD
        # running log methods on ONE SAME THREAD will ensure lockless race-free work
        self.__pool_list = []
        for i in range(4):  # TODO: config!
            self.__pool_list.append(ThreadPoolExecutor(max_workers=1))

    def _get_next_event_id(self):
        """
        get and inc counter
        """
        eid = self.__global_next_event_id
        self.__global_next_event_id += 1
        return eid

    # scheduler component impl

    def _main_task(self):
        return self.main_task()

    def _my_sleep(self):
        pass

    def _my_wake(self):
        pass

    def sleep(self):  # we don't have concept of sleep here
        pass

    def wake(self):  # we don't have concept of sleep here
        pass

    # housekeeping

    # main task

    async def main_task(self):
        self.__logger.info('ui state accessor started')
        self._main_task_is_ready_now()
        await asyncio.gather(self.house_keeping(), self.process_event_queue())
        self.__logger.info('ui state accessor finished')

    async def house_keeping(self):
        stop_task = asyncio.create_task(self._stop_event.wait())
        wakeup_task = None

        while not self._stop_event.is_set():

            self.prune_event_subscriptions()
            if len(self.__pruning_tasks) > 0:  # prune the pruners
                pruned_tasks = []
                for task in self.__pruning_tasks:
                    if task.done():
                        await task
                    else:
                        pruned_tasks.append(task)
                self.__pruning_tasks = pruned_tasks

            ##

            if wakeup_task is None:
                wakeup_task = asyncio.create_task(self._poke_event.wait())
            sleeping_tasks = (stop_task, wakeup_task)

            done, _ = await asyncio.wait(sleeping_tasks, timeout=self.__housekeeping_interval, return_when=asyncio.FIRST_COMPLETED)
            if wakeup_task in done:
                self._reset_poke_event()
                wakeup_task = None

            # end when stop is set
            if stop_task in done:
                break

        # cleaning up potentially pending tasks
        if not stop_task.done():
            stop_task.cancel()
        if wakeup_task is not None and not wakeup_task.done():
            wakeup_task.cancel()

    async def process_event_queue(self):
        stop_task = asyncio.create_task(self._stop_event.wait())
        while not self._stop_event.is_set():
            get_task = asyncio.create_task(self.__task_event_preprocess_queue.get())

            waiting_tasks = (get_task, stop_task)
            done, _ = await asyncio.wait(waiting_tasks, return_when=asyncio.FIRST_COMPLETED)
            if stop_task in done:
                for task in waiting_tasks:
                    if not task.done():
                        task.cancel()
                break

            queue_event_type, event_id, event_timestamp, element_data = await get_task
            if len(self.__task_group_event_logs) == 0:
                continue

            # only in case of removed - we need to know previous groups
            if queue_event_type == QueueEventType.REMOVED:
                _, groups = element_data
                group_sets = (set(groups),)
            elif queue_event_type in (QueueEventType.REMOVED, QueueEventType.ADDED):
                task_datas, groups = element_data  # type: List[TaskData], Optional[Set[str]]
                if groups is None:
                    group_sets = [x.groups for x in task_datas]
                    # group_sets = await self._get_tasks_groups([x.id for x in task_datas])
                else:
                    group_sets = (set(groups),)
            else:  # otherwise we take fresh groups
                group_sets = await self._get_tasks_groups([x.id for x in element_data])

            affected_logs = set()
            for group_set in group_sets:
                for group in group_set:
                    if group not in self.__task_group_to_logs:
                        continue
                    affected_logs.update(x.event_log for x in self.__task_group_to_logs[group])
            if len(affected_logs) == 0:
                continue

            dbuid = self.__data_access.db_uid
            # time to fetch info and build event
            if queue_event_type == QueueEventType.REMOVED:
                task_ids, _ = element_data
                event = TasksRemoved(dbuid, task_ids)
            elif queue_event_type == QueueEventType.ADDED:
                task_datas, _ = element_data
                event = TasksUpdated(dbuid, TaskBatchData(dbuid, {x.id: x for x in task_datas}))
            elif queue_event_type == QueueEventType.UPDATED:
                event = TasksChanged(dbuid, element_data)
            else:
                raise NotImplementedError('IMPOSSIBRU!')

            # force given event_id, ensure correct ids and timestamps
            event.event_id = event_id
            event.timestamp = event_timestamp
            #  put event into logs
            for log in affected_logs:
                executor = self.__get_executor_for_log(log)
                await asyncio.get_event_loop().run_in_executor(executor, log.add_event, event)

    def prune_event_subscriptions(self):
        for k, v in list(self.__task_group_event_logs.items()):  # type: Tuple[Tuple[str, ...], bool], LogSubscription
            if v.is_expired():
                self.remove_task_event_subscription_if_expired(k)

    #

    #

    @property
    def graph_update_id(self):
        return self.__latest_graph_ui_event_id

    def bump_graph_update_id(self):
        self.__latest_graph_ui_state = None
        self.__latest_graph_ui_event_id += 1

    # @atimeit(0.005)
    # async def get_full_ui_state(self, task_groups: Optional[Iterable[str]] = None, skip_dead=True, skip_archived_groups=True) -> UiData:
    #     self.__logger.debug(f'full update for {task_groups}')
    #     now = datetime.now()
    #     group_totals_update_interval = 5
    #     async with self.__data_access.data_connection() as con:
    #         con.row_factory = aiosqlite.Row
    #         async with con.execute('SELECT "id", "type", "name" FROM "nodes"') as cur:
    #             all_nodes = {x['id']: dict(x) for x in await cur.fetchall()}
    #         async with con.execute('SELECT * FROM "node_connections"') as cur:
    #             all_conns = {x['id']: dict(x) for x in await cur.fetchall()}
    #         if not task_groups:  # None or []
    #             all_tasks = dict()
    #             # async with con.execute('SELECT tasks.*, task_splits.origin_task_id, task_splits.split_id, GROUP_CONCAT(task_groups."group") as groups, invocations.progress '
    #             #                        'FROM "tasks" '
    #             #                        'LEFT JOIN "task_splits" ON tasks.id=task_splits.task_id AND tasks.split_level=task_splits.split_level '
    #             #                        'LEFT JOIN "task_groups" ON tasks.id=task_groups.task_id '
    #             #                        'LEFT JOIN "invocations" ON tasks.id=invocations.task_id AND invocations.state = %d '
    #             #                        'GROUP BY tasks."id"' % InvocationState.IN_PROGRESS.value) as cur:
    #             #     all_tasks_rows = await cur.fetchall()
    #             # for task_row in all_tasks_rows:
    #             #     task = dict(task_row)
    #             #     if task['groups'] is None:
    #             #         task['groups'] = set()
    #             #     else:
    #             #         task['groups'] = set(task['groups'].split(','))  # TODO: enforce no commas (,) in group names
    #             #     all_tasks[task['id']] = task
    #         else:
    #             all_tasks = dict()
    #             for group in task_groups:
    #                 # _dbg = time.perf_counter()
    #                 async with con.execute('SELECT tasks.id, tasks.parent_id, tasks.children_count, tasks.active_children_count, tasks.state, tasks.state_details, tasks.paused, tasks.node_id, '
    #                                        'tasks.node_input_name, tasks.node_output_name, tasks.name, tasks.split_level, tasks.work_data_invocation_attempt, '
    #                                        'task_splits.origin_task_id, task_splits.split_id, invocations."id" as invoc_id, GROUP_CONCAT(task_groups."group") as groups '
    #                                        'FROM "tasks" '
    #                                        'LEFT JOIN "task_groups" ON tasks.id=task_groups.task_id AND task_groups."group" == ?'
    #                                        'LEFT JOIN "task_splits" ON tasks.id=task_splits.task_id '
    #                                        'LEFT JOIN "invocations" ON tasks.id=invocations.task_id AND invocations.state = ? '
    #                                        'WHERE task_groups."group" == ? AND tasks.dead {dodead} '
    #                                        'GROUP BY tasks."id"'.format(dodead=f'== 0' if skip_dead else 'IN (0,1)'),
    #                                        (group, InvocationState.IN_PROGRESS.value, group)) as cur:  # NOTE: if you change = to LIKE - make sure to GROUP_CONCAT groups too
    #                     grp_tasks = await cur.fetchall()
    #                 # print(f'fetch groups: {time.perf_counter() - _dbg}')
    #                 for task_row in grp_tasks:
    #                     task = dict(task_row)
    #                     task['progress'] = self.__data_access.mem_cache_invocations.get(task['invoc_id'], {}).get('progress', None)
    #                     task['groups'] = set(task['groups'].split(','))
    #                     if task['id'] in all_tasks:
    #                         all_tasks[task['id']]['groups'].update(task['groups'])
    #                     else:
    #                         all_tasks[task['id']] = task
    #         # _dbg = time.perf_counter()
    #         #async with con.execute('SELECT DISTINCT task_groups."group", task_group_attributes.ctime FROM task_groups LEFT JOIN task_group_attributes ON task_groups."group" = task_group_attributes."group"') as cur:
    #
    #         # some things are updated onlt once in a while, not on every update
    #         need_group_totals_update = (now - (self.__ui_cache.get('last_update_time', None) or datetime.fromtimestamp(0))).total_seconds() > group_totals_update_interval
    #         #async with con.execute('SELECT "group", "ctime", "state", "priority" FROM task_group_attributes' + (f' WHERE state == {TaskGroupArchivedState.NOT_ARCHIVED.value}' if skip_archived_groups else '')) as cur:
    #         if need_group_totals_update:
    #             sqlexpr =  'SELECT "group", "ctime", "state", "priority", tdone, tprog, terr, tall FROM task_group_attributes ' \
    #                        'LEFT JOIN ' \
    #                       f'(SELECT SUM(state=={TaskState.DONE.value}) as tdone, ' \
    #                        f'       SUM(state=={TaskState.IN_PROGRESS.value}) as tprog, ' \
    #                        f'       SUM(state=={TaskState.ERROR.value}) as terr, ' \
    #                        f'       COUNT() as tall, "group" as grp FROM tasks JOIN task_groups ON tasks."id"==task_groups.task_id WHERE tasks.dead==0 GROUP BY "group") ' \
    #                        'ON "grp"==task_group_attributes."group" ' \
    #                        + (f' WHERE state == {TaskGroupArchivedState.NOT_ARCHIVED.value}' if skip_archived_groups else '')
    #         else:
    #             sqlexpr = 'SELECT "group", "ctime", "state", "priority" FROM task_group_attributes' + (f' WHERE state == {TaskGroupArchivedState.NOT_ARCHIVED.value}' if skip_archived_groups else '')
    #         async with con.execute(sqlexpr) as cur:
    #             all_task_groups = {x['group']: dict(x) for x in await cur.fetchall()}
    #         if need_group_totals_update:
    #             self.__ui_cache['last_update_time'] = now
    #             self.__ui_cache['groups'] = {group: {k: attrs[k] for k in ('tdone', 'tprog', 'terr', 'tall')} for group, attrs in all_task_groups.items()}
    #         else:
    #             for group in all_task_groups:
    #                 all_task_groups[group].update(self.__ui_cache['groups'].get(group, {}))
    #
    #         # print(f'distinct groups: {time.perf_counter() - _dbg}')
    #         # _dbg = time.perf_counter()
    #         async with con.execute('SELECT workers."id", '
    #                                'cpu_count, '
    #                                'total_cpu_count, '
    #                                'cpu_mem, '
    #                                'total_cpu_mem, '
    #                                'gpu_count, '
    #                                'total_gpu_count, '
    #                                'gpu_mem, '
    #                                'total_gpu_mem, '
    #                                'workers."hwid", '
    #                                'last_address, workers."state", worker_type, invocations.node_id, invocations.task_id, invocations."id" as invoc_id, '
    #                                'GROUP_CONCAT(worker_groups."group") as groups '
    #                                'FROM workers '
    #                                'LEFT JOIN invocations ON workers."id" == invocations.worker_id AND invocations."state" == 0 '
    #                                'LEFT JOIN worker_groups ON workers."hwid" == worker_groups.worker_hwid '
    #                                'LEFT JOIN resources ON workers.hwid == resources.hwid '
    #                                'GROUP BY workers."id"') as cur:
    #             all_workers = {x['id']: x for x in ({**dict(x),
    #                                                  'last_seen': self.__data_access.mem_cache_workers_state[x['id']]['last_seen'],
    #                                                  'progress': self.__data_access.mem_cache_invocations.get(x['invoc_id'], {}).get('progress', None)
    #                                                  } for x in await cur.fetchall())}
    #             for worker_data in all_workers.values():
    #                 worker_data['groups'] = set(worker_data['groups'].split(',')) if worker_data['groups'] else set()
    #         # print(f'workers: {time.perf_counter() - _dbg}')
    #     return await asyncio.get_event_loop().run_in_executor(None, _create_uidata_from_raw_noasync,
    #                                                           self.__data_access.db_uid,
    #                                                           all_nodes,
    #                                                           all_conns,
    #                                                           all_tasks,
    #                                                           all_workers,
    #                                                           all_task_groups)

    # ui query functions

    async def get_task_groups_ui_state(self, fetch_statistics=False, skip_archived_groups=True, offset=0, limit=-1) -> TaskGroupBatchData:
        self.__logger.debug(f'tasks groups update for offset={offset}, limit={"unlim" if limit < 0 else limit}')
        # group_totals_update_interval = 5
        # now = datetime.now()
        async with self.__data_access.data_connection() as con, \
                aperformance_measurer(threshold_to_report=0.005, name='get_task_groups_ui_state'):
            con.row_factory = aiosqlite.Row
            # need_group_totals_update = (now - (self.__ui_cache.get('last_update_time', None) or datetime.fromtimestamp(0))).total_seconds() > group_totals_update_interval
            # fetch_statistics = fetch_statistics and need_group_totals_update
            if fetch_statistics:
                sqlexpr = 'SELECT "group", "ctime", "state", "priority", tdone, tprog, terr, tall FROM task_group_attributes ' \
                          'LEFT JOIN ' \
                          f'(SELECT SUM(state=={TaskState.DONE.value}) as tdone, ' \
                          f'       SUM(state=={TaskState.IN_PROGRESS.value}) as tprog, ' \
                          f'       SUM(state=={TaskState.ERROR.value}) as terr, ' \
                          f'       COUNT() as tall, "group" as grp FROM tasks JOIN task_groups ON tasks."id"==task_groups.task_id WHERE tasks.dead==0 GROUP BY "group") ' \
                          'ON "grp"==task_group_attributes."group" ' \
                          + (f' WHERE state == {TaskGroupArchivedState.NOT_ARCHIVED.value}' if skip_archived_groups else '')
            else:
                sqlexpr = 'SELECT "group", "ctime", "state", "priority" FROM task_group_attributes' + (f' WHERE state == {TaskGroupArchivedState.NOT_ARCHIVED.value}' if skip_archived_groups else '')
            if offset > 0 or limit >= 0:
                sqlexpr += f' LIMIT {int(limit)} OFFSET {int(offset)}'
            async with con.execute(sqlexpr) as cur:
                all_task_groups = {x['group']: dict(x) for x in await cur.fetchall()}

            # # now update cache or pull statistics from cache
            # if fetch_statistics:
            #     self.__ui_cache['last_update_time'] = now
            #     self.__ui_cache['groups'] = {group: {k: attrs[k] for k in ('tdone', 'tprog', 'terr', 'tall')} for group, attrs in all_task_groups.items()}
            # elif fetch_statistics:  # if fetch requested, but cache is still valid - get cache
            #     for group in all_task_groups:
            #         all_task_groups[group].update(self.__ui_cache['groups'].get(group, {}))

        return await asyncio.get_event_loop().run_in_executor(None, _pack_task_groups, self.__data_access.db_uid, all_task_groups)

    async def get_tasks_ui_state(self, task_groups: Optional[Iterable[str]] = None, skip_dead=True) -> TaskBatchData:
        self.__logger.debug(f'tasks update for {task_groups}')
        async with self.__data_access.data_connection() as con, \
                aperformance_measurer(threshold_to_report=0.005, name='get_tasks_ui_state'):
            con.row_factory = aiosqlite.Row

            all_tasks = dict()
            for group in task_groups:
                async with con.execute('SELECT tasks.id, tasks.parent_id, tasks.children_count, tasks.active_children_count, tasks.state, tasks.state_details, tasks.paused, tasks.node_id, '
                                       'tasks.node_input_name, tasks.node_output_name, tasks.name, tasks.split_level, tasks.work_data_invocation_attempt, '
                                       'task_splits.origin_task_id, task_splits.split_id, invocations."id" as invoc_id, GROUP_CONCAT(task_groups."group") as groups '
                                       'FROM "tasks" '
                                       'LEFT JOIN "task_groups" ON tasks.id=task_groups.task_id AND task_groups."group" == ?'
                                       'LEFT JOIN "task_splits" ON tasks.id=task_splits.task_id '
                                       'LEFT JOIN "invocations" ON tasks.id=invocations.task_id AND invocations.state = ? '
                                       'WHERE task_groups."group" == ? AND tasks.dead {dodead} '
                                       'GROUP BY tasks."id"'.format(dodead=f'== 0' if skip_dead else 'IN (0,1)'),
                                       (group, InvocationState.IN_PROGRESS.value, group)) as cur:  # NOTE: if you change = to LIKE - make sure to GROUP_CONCAT groups too
                    grp_tasks = await cur.fetchall()
                # print(f'fetch groups: {time.perf_counter() - _dbg}')
                for task_row in grp_tasks:
                    task = dict(task_row)
                    task['progress'] = self.__data_access.mem_cache_invocations.get(task['invoc_id'], {}).get('progress', None)
                    task['groups'] = set(task['groups'].split(','))
                    if task['id'] in all_tasks:
                        all_tasks[task['id']]['groups'].update(task['groups'])
                    else:
                        all_tasks[task['id']] = task

        return await asyncio.get_event_loop().run_in_executor(None, _pack_tasks_data, self.__data_access.db_uid, all_tasks)

    async def get_nodes_ui_state(self) -> NodeGraphStructureData:
        if self.__latest_graph_ui_state is None:
            self.__logger.debug('nodes update')
            async with self.__data_access.data_connection() as con, \
                    aperformance_measurer(threshold_to_report=0.005, name='get_nodes_ui_state'):
                con.row_factory = aiosqlite.Row
                async with con.execute('SELECT "id", "type", "name" FROM "nodes"') as cur:
                    all_nodes = {x['id']: dict(x) for x in await cur.fetchall()}
                async with con.execute('SELECT * FROM "node_connections"') as cur:
                    all_conns = {x['id']: dict(x) for x in await cur.fetchall()}

            self.__latest_graph_ui_state = await asyncio.get_event_loop().run_in_executor(None, _pack_nodes_connections_data, self.__data_access.db_uid, all_nodes, all_conns)
        return self.__latest_graph_ui_state

    async def get_workers_ui_state(self) -> WorkerBatchData:
        self.__logger.debug('workers update')
        async with self.__data_access.data_connection() as con, \
                aperformance_measurer(threshold_to_report=0.005, name='get_workers_ui_state'):
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT workers."id", '
                                   'cpu_count, '
                                   'total_cpu_count, '
                                   'cpu_mem, '
                                   'total_cpu_mem, '
                                   'gpu_count, '
                                   'total_gpu_count, '
                                   'gpu_mem, '
                                   'total_gpu_mem, '
                                   'workers."hwid", '
                                   'last_address, workers."state", worker_type, invocations.node_id, invocations.task_id, invocations."id" as invoc_id, '
                                   'GROUP_CONCAT(worker_groups."group") as groups '
                                   'FROM workers '
                                   'LEFT JOIN invocations ON workers."id" == invocations.worker_id AND invocations."state" == 0 '
                                   'LEFT JOIN worker_groups ON workers."hwid" == worker_groups.worker_hwid '
                                   'LEFT JOIN resources ON workers.hwid == resources.hwid '
                                   'GROUP BY workers."id"') as cur:
                all_workers = {x['id']: x for x in ({**dict(x),
                                                     'last_seen': self.__data_access.mem_cache_workers_state[x['id']]['last_seen'],
                                                     'progress': self.__data_access.mem_cache_invocations.get(x['invoc_id'], {}).get('progress', None)
                                                     } for x in await cur.fetchall())}
                for worker_data in all_workers.values():
                    worker_data['groups'] = set(worker_data['groups'].split(',')) if worker_data['groups'] else set()

        return await asyncio.get_event_loop().run_in_executor(None, _pack_workers_from_raw, self.__data_access.db_uid, all_workers)

    #
    # task group mapping related crap
    #

    def force_refresh_task_group_mapping(self):
        """
        call this if you know task groups changed.
         you DON'T NEED TO call this if tasks were added or removed together with their groups -
         because calls to scheduler_reports_task_added scheduler_reports_task_removed_from_group will call this method anyway
        """
        self.__requested_task_group_force_refresh = True
        self.__task_group_mapping = None
        self.__group_task_mapping = None

    async def __refetch_groups(self):
        """
        refetch __task_group_mapping from database
        """
        while self.__requested_task_group_force_refresh:  # if it was set again during awaits inside
            self.__requested_task_group_force_refresh = False
            async with self.__task_group_mapping_update_alock:
                async with self.__data_access.data_connection() as con, \
                        aperformance_measurer(threshold_to_report=0.005, name='get_workers_ui_state'):
                    con.row_factory = aiosqlite.Row
                    async with con.execute('SELECT task_id, "group" FROM task_groups') as cur:
                        rows = await cur.fetchall()

                def _do():
                    d = {}
                    di = {}
                    for row in rows:
                        d.setdefault(row['task_id'], set()).add(row['group'])
                        di.setdefault(row['group'], set()).add(row['task_id'])
                    return d, di

                self.__task_group_mapping, self.__group_task_mapping = await asyncio.get_event_loop().run_in_executor(None, _do)

    async def _get_tasks_groups(self, task_ids: List[int]) -> List[Set[str]]:
        if self.__task_group_mapping is None:
            await self.__refetch_groups()
        ret = []
        for task_id in task_ids:
            ret.append(self.__task_group_mapping.get(task_id) or set())
        return ret

    async def _get_group_tasks(self, group: str) -> Set[int]:
        if self.__group_task_mapping is None:
            await self.__refetch_groups()
        return self.__group_task_mapping.get(group, None)
    #

    def scheduler_reports_tasks_added(self, task_datas: List[TaskData], added_where: Optional[Set[str]] = None):
        """
        task info will be queried from DB
        """
        self.force_refresh_task_group_mapping()
        if len(self.__task_group_event_logs) == 0:
            return
        eid = self._get_next_event_id()
        ets = time.time_ns()
        self.__task_event_preprocess_queue.put_nowait((QueueEventType.ADDED, eid, ets, (task_datas, added_where)))

    def scheduler_reports_task_added(self, task_data: TaskData, added_where: Optional[Set[str]] = None):
        """
        if task_raw is None - it will be fetched from DB
        """
        self.scheduler_reports_tasks_added([task_data], added_where)

    def scheduler_reports_tasks_updated(self, task_deltas: List[TaskDelta]):
        if len(self.__task_group_event_logs) == 0:
            return
        eid = self._get_next_event_id()
        ets = time.time_ns()
        self.__task_event_preprocess_queue.put_nowait((QueueEventType.UPDATED, eid, ets, task_deltas))

    def scheduler_reports_task_updated(self, task_delta: TaskDelta):
        self.scheduler_reports_tasks_updated([task_delta])

    def scheduler_reports_tasks_removed_from_group(self, task_ids: List[int], groups: Iterable[str]):
        self.force_refresh_task_group_mapping()
        if len(self.__task_group_event_logs) == 0:
            return
        eid = self._get_next_event_id()
        ets = time.time_ns()
        self.__task_event_preprocess_queue.put_nowait((QueueEventType.REMOVED, eid, ets, (task_ids, groups)))

    def scheduler_reports_task_groups_added(self, groups):
        self.force_refresh_task_group_mapping()

    def scheduler_reports_task_groups_changed(self, groups):  # TODO: this has to be changed, for now it's way too broad
        self.force_refresh_task_group_mapping()

    def scheduler_reports_task_groups_removed(self, groups):
        self.force_refresh_task_group_mapping()

    #

    def __get_executor_for_log(self, log) -> ThreadPoolExecutor:
        return self.__pool_list[hash(log) % len(self.__pool_list)]

    async def subscribe_to_task_events_for_groups(self, task_groups: Iterable[str], skip_dead: bool, subscribe_for_seconds: float) -> Tuple[TaskEvent, ...]:
        """

        :param task_groups:
        :param skip_dead:
        :param subscribe_for_seconds:
        :return:
        """
        last_race_check_id = -1
        group_key = (tuple(sorted(task_groups)), skip_dead)
        if group_key in self.__task_group_event_logs:  # if so - update
            self.__task_group_event_logs[group_key].expiration_timestamp = time.time() + subscribe_for_seconds
            self.__logger.debug(f'ui subscription to "{group_key}" was prolonged')

            log = self.__task_group_event_logs[group_key].event_log
            executor = self.__get_executor_for_log(log)
            events = await asyncio.get_event_loop().run_in_executor(executor, log.get_since_event, -1, True)  # try to see if we already have a chain of events starting with full state update

            if len(events) > 0 and events[0].event_type == UIEventType.FULL_STATE:  # so if we have a set of events starting with full state
                return events
            self.__logger.debug('no FULL_STATE event in log, reissuing full ui update event')
            if len(events) > 0:
                last_race_check_id = events[-1].event_id
        else:
            log_sub = LogSubscription(time.time() + subscribe_for_seconds, SchedulerEventLog(log_time_length_max=60))
            self.__task_group_event_logs[group_key] = log_sub
            for group in task_groups:
                self.__task_group_to_logs.setdefault(group, []).append(log_sub)

            # now add pruning task
            async def _timed_prune():
                while (log_sub := self.__task_group_event_logs.get(group_key)) is not None:
                    await asyncio.sleep((log_sub.expiration_timestamp - time.time()) * 1.1)  # 1.1 just cuz
                    self.remove_task_event_subscription_if_expired(group_key)

            self.__pruning_tasks.append(asyncio.create_task(_timed_prune()))
            self.__logger.debug(f'ui subscription to "{group_key}" was added')

        log = self.__task_group_event_logs[group_key].event_log
        executor = self.__get_executor_for_log(log)

        # there's no way i can think of to guarantee no race conditions between full state query and other events happening
        # in such case we cannot order full update properly, no way to tell what happened first
        # so for now we just detect (it's not even a guaranteed detect...) if other events happen during full update
        eid = self._get_next_event_id()  # it seeeems that this will be the least destructive way. we risk having older events after full update, but at least final picture will be correct
        ets = time.time_ns()
        state = await self.get_tasks_ui_state(task_groups, skip_dead)
        planck_events = log.get_since_event(last_race_check_id)  # no need for executor - this should be fast, and we don't want to poke event loop yet
        if len(planck_events) > 0:
            self.__logger.warning(f'full state race condition detected! could not order events: {eid}, {[e.event_id for e in planck_events]}')

        state_event = TaskFullState(self.__data_access.db_uid, state)
        state_event.event_id = eid
        state_event.timestamp = ets
        # even though there might be older events in the queue - since we have strict global event id - the state_event will be put into a proper place
        await asyncio.get_event_loop().run_in_executor(executor, log.add_event, state_event)

        return state_event,

    def has_task_event_subscription(self, key: Tuple[Tuple[str, ...], bool]) -> bool:
        return key in self.__task_group_event_logs

    def remove_task_event_subscription_if_expired(self, key: Tuple[Tuple[str, ...], bool]) -> bool:
        """
        removes subscription key ONLY if it's expired
        returns True if subscription was indeed removed
        """
        if (log_sub := self.__task_group_event_logs.get(key)) and log_sub is not None and log_sub.is_expired():
            self.__task_group_event_logs.pop(key)
            for group in key[0]:
                self.__task_group_to_logs[group].remove(log_sub)
                if len(self.__task_group_to_logs[group]) == 0:
                    self.__task_group_to_logs.pop(group)
            self.__logger.debug(f'ui subscription to "{key}" was removed')
            return True
        return False

    async def get_events_for_groups_since_event_id(self, task_groups: Iterable[str], skip_dead: bool, last_known_event_id: int) -> Tuple[TaskEvent, ...]:
        group_key = (tuple(sorted(task_groups)), skip_dead)
        if group_key not in self.__task_group_event_logs:
            raise NotSubscribedError()
        log = self.__task_group_event_logs[group_key].event_log
        executor = self.__get_executor_for_log(log)
        events = await asyncio.get_event_loop().run_in_executor(executor, log.get_since_event, last_known_event_id, True)
        return events

    def subscriptions_count(self):
        return len(self.__task_group_event_logs)

# scheduler helpers

def _pack_workers_from_raw(db_uid: int, ui_workers: dict) -> "WorkerBatchData":
    """
    this is scheduler helper function, it's incoming data format is dictated purely by scheduler
    """
    workers = {}
    for worker_id, worker_raw in ui_workers.items():
        assert worker_id == worker_raw['id']
        res = WorkerResources(worker_raw['cpu_count'], worker_raw['total_cpu_count'],
                              worker_raw['cpu_mem'], worker_raw['total_cpu_mem'],
                              worker_raw['gpu_count'], worker_raw['total_gpu_count'],
                              worker_raw['gpu_mem'], worker_raw['total_gpu_mem'])
        workers[worker_id] = WorkerData(worker_id, res, str(worker_raw['hwid']), worker_raw['last_address'], worker_raw['last_seen'],
                                        WorkerState(worker_raw['state']), WorkerType(worker_raw['worker_type']),
                                        worker_raw['node_id'], worker_raw['task_id'], worker_raw['invoc_id'], worker_raw['progress'],
                                        worker_raw['groups'])

    return WorkerBatchData(db_uid, workers)


def _pack_nodes_connections_data(db_uid: int, ui_nodes, ui_connections) -> "NodeGraphStructureData":
    if ui_nodes is None or ui_connections is None:
        if ui_connections is not None or ui_connections is not None:
            raise RuntimeError('both ui_nodes and ui_connections must be none, or not none')
    nodes = {}
    connections = {}
    for node_id, node_raw in ui_nodes.items():
        assert node_id == node_raw['id']
        node_data = NodeData(node_id, node_raw['name'], node_raw['type'])
        nodes[node_id] = node_data
    for conn_id, con_raw in ui_connections.items():
        assert conn_id == con_raw['id']
        conn_data = NodeConnectionData(conn_id, con_raw['node_id_in'], con_raw['in_name'], con_raw['node_id_out'], con_raw['out_name'])
        connections[conn_id] = conn_data

    return NodeGraphStructureData(db_uid, nodes, connections)


def _pack_task_data(task_id, task_raw: dict) -> "TaskData":
    return TaskData(task_id, task_raw['parent_id'], task_raw['children_count'], task_raw['active_children_count'],
           TaskState(task_raw['state']), task_raw['state_details'], task_raw['paused'] != 0, task_raw['node_id'],
           task_raw['node_input_name'], task_raw['node_output_name'], task_raw['name'], task_raw['split_level'],
           task_raw['work_data_invocation_attempt'], task_raw['progress'], task_raw['origin_task_id'],
           task_raw['split_id'], task_raw['invoc_id'], task_raw['groups'])


def _pack_tasks_data(db_uid: int, ui_tasks: Union[dict, List[dict]]) -> "TaskBatchData":
    tasks = {}
    if isinstance(ui_tasks, dict):
        for task_id, task_raw in ui_tasks.items():
            assert task_id == task_raw['id']
            task_data = _pack_task_data(task_id, task_raw)
            tasks[task_id] = task_data
    elif isinstance(ui_tasks, list):
        for task_raw in ui_tasks:
            task_id = task_raw['id']
            task_data = _pack_task_data(task_id, task_raw)
            tasks[task_id] = task_data

    return TaskBatchData(db_uid, tasks)


def _pack_task_groups(db_uid: int, all_task_groups) -> "TaskGroupBatchData":
    task_groups = {}
    for group_name, group_raw in all_task_groups.items():
        assert group_name == group_raw['group']
        if 'tdone' in group_raw:  # if has stat:
            stat = TaskGroupStatisticsData(group_raw['tdone'], group_raw['tprog'], group_raw['terr'], group_raw['tall'])
        else:
            stat = None
        task_groups[group_name] = TaskGroupData(group_name, group_raw['ctime'], TaskGroupArchivedState(group_raw['state']),
                                                group_raw['priority'], stat)

    return TaskGroupBatchData(db_uid, task_groups)


def _create_uidata_from_raw_noasync(db_uid, ui_nodes, ui_connections, ui_tasks, ui_workers, all_task_groups):

    node_graph_data = _pack_nodes_connections_data(db_uid, ui_nodes, ui_connections)
    tasks = _pack_tasks_data(db_uid, ui_tasks)
    worker_batch_data = _pack_workers_from_raw(db_uid, ui_workers)
    task_groups = _pack_task_groups(db_uid, all_task_groups)

    return UiData(db_uid, node_graph_data, tasks, worker_batch_data, task_groups)



import sys
import argparse
from . import logging


def main(argv):
    parser = argparse.ArgumentParser('lifeblood', description='task execution thingie')
    parser.add_argument('--loglevel', help='logging level, like DEBUG, INFO, WARNING, ERROR')

    subparsers = parser.add_subparsers(title='command', required=True, dest='command')
    schedparser = subparsers.add_parser('scheduler', description='run main scheduler server', add_help=False)

    workerparser = subparsers.add_parser('worker', description='run a worker', add_help=False)
    workerparser.add_argument('args', nargs=argparse.REMAINDER, help='arguments to pass to the worker')

    viewparser = subparsers.add_parser('viewer', description='run default viewer')

    poolparser = subparsers.add_parser('pool', description='run a worker pool', add_help=False)
    poolparser.add_argument('args', nargs=argparse.REMAINDER, help='arguments to pass to the pool')

    envresparser = subparsers.add_parser('resolver', description='configurate environment resolvers', add_help=False)
    envresparser.add_argument('args', nargs=argparse.REMAINDER, help='arguments to pass to the pool')

    baseopts, _ = parser.parse_known_args(argv)
    cmd_id = argv.index(baseopts.command)
    cmd_argv = argv[cmd_id + 1:]
    baseargv = argv[:cmd_id + 1]
    opts = parser.parse_args(baseargv)

    if opts.loglevel is not None:
        logging.set_default_loglevel(opts.loglevel)

    if opts.command == 'scheduler':
        from .scheduler import main
        return main(cmd_argv)
    elif opts.command == 'worker':
        from .worker import main
        return main(cmd_argv)
    elif opts.command == 'pool':
        from .worker_pool import main
        return main(cmd_argv)
    elif opts.command == 'viewer':
        try:
            from lifeblood_viewer.launch import main
        except ImportError as e:
            logger = logging.get_logger('main')
            logger.error('Viewer python package not found. In needs be installed separately with smth like "pip install lifeblood_viewer"')
            logger.exception(e)
            return
        return main(cmd_argv)
    elif opts.command == 'resolver':
        from .environment_resolver import main
        return main(cmd_argv)


def console_entry_point():
    sys.exit(main(sys.argv[1:]))


if __name__ == '__main__':
    console_entry_point()



#!/usr/bin/env python

import sys
import argparse
import random
import json
import re
import lifeblood_connection
from pathlib import Path
from http.server import ThreadingHTTPServer, BaseHTTPRequestHandler
from cgi import parse_multipart

from typing import Dict, List, Optional, Tuple


class DataAccessor:
    def has_data(self, key: str):
        raise NotImplementedError()

    def get_data(self, key: str) -> str:
        raise NotImplementedError()

    def set_data(self, key: str, data: str):
        raise NotImplementedError()

    def store_input_files(self, identifier: str, contents: List[bytes]) -> Tuple[str]:
        raise NotImplementedError()

    def set_output_files_future(self, identifier: str, future):
        raise NotImplementedError()

    def output_files_ready(self, identifier: str) -> bool:
        raise NotImplementedError()

    def get_output_files(self, identifier: str) -> Dict[str, bytes]:
        raise NotImplementedError()


class DictDataAccessor(DataAccessor):
    def __init__(self, storage_base_path: Path):
        self.__dict: Dict[str, str] = {}
        self.__futures = {}
        self.__results = {}
        self.__input_storage_base = storage_base_path

    def has_data(self, key: str):
        return key in self.__dict

    def get_data(self, key: str) -> str:
        return self.__dict[key]

    def set_data(self, key: str, data: str):
        self.__dict[key] = data

    def store_input_files(self, identifier: str, contentss: List[bytes]) -> Tuple[str]:
        if identifier is None:
            raise NotImplementedError()
        base_file_path = self.__input_storage_base / identifier
        base_file_path.mkdir(exist_ok=True, parents=True)
        file_paths = []
        for i, contents in enumerate(contentss):
            file_path = base_file_path / f'file_{i}'
            with open(file_path, 'wb') as f:
                f.write(contents)
            file_paths.append(str(file_path))
        return tuple(file_paths)

    def set_output_files_future(self, identifier: str, future):
        if identifier in self.__futures:
            raise RuntimeError(f'identifier {identifier} already has a future')
        self.__futures[identifier] = future

    def output_files_ready(self, identifier: str) -> bool:
        if identifier in self.__results:
            return True
        if identifier in self.__futures:
            return self.__futures[identifier].is_done()
        raise ValueError(f'identifier {identifier} is neither ready nor awaited')

    def get_output_files(self, identifier: str) -> Dict[str, bytes]:
        if identifier in self.__results:
            return self.__read_files(self.__results[identifier])
        if identifier in self.__futures:
            iid, raw_reply = self.__futures[identifier].wait()
            reply = json.loads(raw_reply.decode('UTF-8'))
            self.__results[identifier] = tuple(reply.get('files', ()))
            self.__futures.pop(identifier)
            return self.__read_files(self.__results[identifier])
        raise ValueError(f'identifier {identifier} is neither ready nor awaited')

    @classmethod
    def __read_files(cls, file_paths: Tuple[str]) -> Dict[str, bytes]:
        datas = {}
        for file_path in file_paths:
            with open(file_path, 'rb') as f:
                datas[file_path] = f.read()
        return datas


class Handler(BaseHTTPRequestHandler):
    def __init__(self, main_html: str, result_not_ready_html: str, result_ready_html: str, css: str, data_accessor: DataAccessor, expect_reply: bool, *args, **kwargs):
        self.__index_html = main_html
        self.__result_not_ready_html = result_not_ready_html
        self.__result_ready_html = result_ready_html
        self.__css = css
        self.__data_accessor = data_accessor
        self.__expect_reply = expect_reply
        super().__init__(*args, **kwargs)

    def _respond_not_found(self):
        self.send_response(404)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(b'<h1>YOU SHALL NOT PASS</h1>')

    def _respond_redirect(self, to: str):
        self.send_response(301)
        self.send_header('Location', to)
        self.end_headers()

    def do_GET(self):
        if self.path == '/':
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(self.__index_html.encode('utf-8'))
        elif self.path == '/styles.css':
            self.send_response(200)
            self.send_header("Content-type", "text/css")
            self.end_headers()
            self.wfile.write(self.__css.encode('utf-8'))
        elif self.path.startswith('/result/'):
            # if no expecting reply - just show empty result page
            if not self.__expect_reply:
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                body = self.__result_ready_html.format(result='')
                self.end_headers()
                self.wfile.write(body.encode('utf-8'))
                return
            #
            # otherwise we wait for result to be ready
            key = self.path[len('/result/'):]
            file_key = None
            if key.count('/') > 1:
                self._respond_not_found()
                return
            elif '/' in key:
                key, file_key = key.split('/', 1)

            try:
                result_ready = self.__data_accessor.output_files_ready(key)
            except ValueError:
                self._respond_not_found()
                return
            if result_ready:
                files = self.__data_accessor.get_output_files(key)
                keys_ordered = tuple(files.keys())
                self.send_response(200)
                if file_key is None:
                    self.send_header("Content-type", "text/html")
                    body = self.__result_ready_html.format(
                        result='\n'.join(f'<br><a href="/result/{key}/{i}" download="{Path(path).name}">{Path(path).name}</a>' for i, path in enumerate(keys_ordered))
                    )
                    self.end_headers()
                    self.wfile.write(body.encode('utf-8'))
                else:
                    # file key is just order in default sorted file dict
                    key_file_path = keys_ordered[int(file_key)]
                    data = files[key_file_path]
                    file_name = Path(key_file_path).name.replace('"', '_')
                    self.send_header('Content-Disposition', f'attachment; filename=\"{file_name}\"')
                    self.send_header('Content-Length', f'{len(data)}')
                    self.end_headers()
                    self.wfile.write(data)

                return
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(self.__result_not_ready_html.format(key=key, rand=random.random()).encode('utf-8'))
        else:
            self._respond_not_found()

    def do_POST(self):
        if self.path != '/do_something_smart':
            self._respond_not_found()
            return

        files = parse_multipart(self.rfile, {k.strip(): v.strip().encode('utf-8') for part in
                                             self.headers['Content-Type'].split(';')[1:] for k, v in
                                             (part.split('=', 1),)})
        images = files.get('images')
        if not images:
            self._respond_not_found()

        key = lifeblood_connection.generate_addressee()

        file_paths = self.__data_accessor.store_input_files(key, images)

        attribs = {
            'files': file_paths,
        }
        if self.__expect_reply:
            attribs.update({
                'server_iid': lifeblood_connection.get_my_invocation_id(),
                'server_addressee': key,
            })

        lifeblood_connection.create_task('file processing', attribs, blocking=True)
        self.__data_accessor.set_output_files_future(key, lifeblood_connection.message_to_invocation_receive(key, blocking=False))

        self._respond_redirect(f'/result/{key}')


def main(args, base_path: Path):
    parser = argparse.ArgumentParser(description='run simplest web server for internal network use')
    parser.add_argument('--port', '-p', type=int, help='port to open server on')
    parser.add_argument('--ip', '-a', type=str, help='address the server will listen to')
    parser.add_argument('--title', type=str, help='page welcome text')
    parser.add_argument('--expect-reply', action='store_true', help='page welcome text')
    parser.add_argument('--base_storage_path', type=str, help='where to store input files')
    opts = parser.parse_args(args)

    # html+css based on https://codepen.io/Scribblerockerz/pen/qdWzJw
    with open(base_path/"index.html") as f:
        html_body = f.read()
    html_body = html_body.format(title=opts.title or '<strong>File upload</strong> example',
                                 header_title=re.sub(r'<([^<>]+)>(.*?)<\/\1>', r'\2', opts.title or 'File upload'))

    with open(base_path/"styles.css") as f:
        css_body = f.read()

    with open(base_path/"result_not_ready.html") as f:
        result_not_ready_html_template = f.read()

    with open(base_path/"result_ready.html") as f:
        result_ready_html_template = f.read()

    data_accessor = DictDataAccessor(Path(opts.base_storage_path))

    server = ThreadingHTTPServer(
        (opts.ip or '127.0.0.1', opts.port),
        lambda *args, **kwargs: Handler(html_body, result_not_ready_html_template, result_ready_html_template, css_body, data_accessor, opts.expect_reply, *args, **kwargs)
    )
    server.serve_forever()


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:], Path(sys.argv[0]).parent) or 0)



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionMeshing


class AlicevisionMeshing(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionMeshing, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Structure From Motion', NodeParameterType.STRING, "`task['av_structure_from_motion']`")
            ui.add_parameter('depthMapsFolder', 'Depth Maps', NodeParameterType.STRING, "`task['av_depth_map_filtered']`")
            ui.add_parameter('outputMesh', 'Output Mesh', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/Meshing/mesh.obj")
            ui.add_parameter('output', 'Output Dense Point Cloud', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/Meshing/densePointCloud.abc")
            ui.add_separator()
            # TODO: add bounding box parameter
            ui.add_parameter('estimateSpaceMinObservationAngle', 'Min Observations Angle For SfM Space Estimation', NodeParameterType.INT, 10)
            ui.add_parameter('maxInputPoints', 'Max Input Points', NodeParameterType.INT, 50_000_000)
            ui.add_parameter('maxPoints', 'Max Points', NodeParameterType.INT, 5_000_000)
            ui.add_parameter('voteFilteringForWeaklySupportedSurfaces', 'Weakly Supported Surface Support', NodeParameterType.BOOL, True)
            ui.add_parameter('colorizeOutput', 'Colorize Output', NodeParameterType.BOOL, False)

    @classmethod
    def label(cls) -> str:
        return 'AV Meshing'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'meshing'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_meshing'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output_mesh = Path(context.param_value('outputMesh').strip())
        if not output_mesh.is_absolute():
            raise ProcessingError('output mesh path must be absolute and not empty')
        output = Path(context.param_value('output').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.parent.mkdir(parents=True, exist_ok=True)
        output_mesh.parent.mkdir(parents=True, exist_ok=True)

        args = ['aliceVision_meshing', '--verboseLevel', 'info']
        args += ['--input', context.param_value('input').strip()]
        args += ['--depthMapsFolder', context.param_value('depthMapsFolder').strip()]
        for param_name in ('estimateSpaceMinObservationAngle', 'maxInputPoints', 'maxPoints',
                           'voteFilteringForWeaklySupportedSurfaces', 'colorizeOutput'):
            args += [f'--{param_name}', context.param_value(param_name)]

        args += ['--output', output]
        args += ['--outputMesh', output_mesh]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_mesh_object', str(output_mesh))
        res.set_attribute('av_dense_point_cloud', str(output))
        return res



import re

from lifeblood.logging import get_logger
from typing import Iterable


def match(match_expr: str, items: Iterable[str]):
    """
    return items matching match_expr
    TODO: speed it up
    :param match_expr:
    :param items:
    :return:
    """
    get_logger('scheduler').warning('this function is deprecated')
    res = []

    parts = []
    escaping = False
    in_quotes = False
    in_part = False
    offstart = 0
    part = ''
    for offset, symbol in enumerate(match_expr):
        if symbol in (' ', '\t'):
            if in_part and not escaping and not in_quotes:
                parts.append(part + match_expr[offstart:offset])
                in_part = False
            escaping = False
        else:
            if not in_part:
                in_part = True
                offstart = offset
                part = ''
            if symbol == '"' and not escaping:
                in_quotes = not in_quotes
                part += match_expr[offstart:offset]
                offstart = offset + 1
            if symbol == '\\' and not escaping:
                escaping = not escaping
                part += match_expr[offstart:offset]
                offstart = offset + 1
            else:
                escaping = False
    if in_part:
        parts.append(part + match_expr[offstart:])

    #
    #

    for item in items:
        good = False
        for part in parts:
            if part == '*':  # TODO: this is a special case
                good = True
            elif part[0] == '^':
                if part[1:] == item:
                    good = False
            else:
                if part == item:
                    good = True

        if good:
            res.append(item)

    return res



import os
import re
from copy import copy, deepcopy
import asyncio
import pickle
from types import MappingProxyType
from .enums import WorkerType
from .net_classes import WorkerResources

from typing import Optional, Iterable, Union, Dict, List, Set, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from .environment_resolver import EnvironmentResolverArguments


class InvocationNotFinished(RuntimeError):
    pass


class BadProgressRegexp(RuntimeError):
    pass


class Environment(dict):
    def __init__(self, *args, **kwargs):
        super(Environment, self).__init__(*args, **kwargs)
        self.__expandre = re.compile(r'\$(?:(\w+)|{(\w+)})')
        self.__extra_expand_dict = {}

    def set_extra_expand_dict(self, extra: dict):
        self.__extra_expand_dict = extra

    def expand(self, value: str) -> str:
        def _onmatch(match):
            key = match.group(1) or match.group(2)
            return self.get(key, self.__extra_expand_dict.get(key, None))
        return self.__expandre.sub(_onmatch, value)

    def __setitem__(self, key: str, value):
        if not isinstance(value, str):
            value = str(value)
        super(Environment, self).__setitem__(key, self.expand(value))

    def prepend(self, key: str, value):
        """
        treat key as path list and prepend to the list
        """
        if key not in self:
            self[key] = value
            return
        if not isinstance(value, str):
            value = str(value)
        self[key] = os.pathsep.join((self.expand(value), self[key]))

    def append(self, key: str, value):
        """
        treat key as path list and append to the list
        """
        if key not in self:
            self[key] = value
            return
        if not isinstance(value, str):
            value = str(value)
        self[key] = os.pathsep.join((self[key], self.expand(value)))


class InvocationEnvironment:
    def __init__(self, *args, **kwargs):
        super(InvocationEnvironment, self).__init__(*args, **kwargs)
        self.__action_queue = []

    def set_variable(self, key: str, value):
        if not isinstance(value, str):
            value = str(value)
        self.__action_queue.append(('__setitem__', key, value))

    def resolve(self, base_env: Optional[Environment] = None, additional_environment_to_expand_with: Optional[Environment] = None) -> Environment:
        """
        resolves action queue and produces final environment
        """
        if base_env is not None:
            env = copy(base_env)
        else:
            env = Environment()
        if additional_environment_to_expand_with is not None:
            env.set_extra_expand_dict(additional_environment_to_expand_with)
        for method, *args in self.__action_queue:
            getattr(env, method)(*args)
        return env

    def _enqueue_kv_method(self, method: str, key: str, value):
        if not isinstance(value, str):
            value = str(value)
        self.__action_queue.append((method, key, value))

    # def __getattr__(self, item):
    #     return lambda k, v: self._enqueue_kv_method(item, k, v)

    # these 2 guys are explicitly added only for IDE popup hints
    def prepend(self, key: str, value):
        self._enqueue_kv_method('prepend', key, value)

    def append(self, key: str, value):
        self._enqueue_kv_method('append', key, value)


class InvocationRequirements:
    """
    requirements a worker has to match in order to be able to pick this task

    logic is that workers must fit into minimums, but up to pref (preferred) amount of resources will be actually taken
    for example, you might want minimum of 1 CPU core, but would prefer to use 4
    then a 1 core machine may pick up one task, but 16 core machine will pick just 4 tasks
    """
    def __init__(self, *,
                 min_cpu_count: Optional[int] = None,
                 min_memory_bytes: Optional[int] = None,
                 min_gpu_count: Optional[int] = None,
                 min_gpu_memory_bytes: Optional[int] = None,
                 groups: Optional[Iterable[str]] = None,
                 worker_type: WorkerType = WorkerType.STANDARD):
        self.__groups = set(groups) if groups is not None else set()
        self.__worker_type = worker_type
        self.__min_cpu_count = min_cpu_count or 0
        self.__min_memory_bytes = min_memory_bytes or 0
        self.__min_gpu_count = min_gpu_count or 0
        self.__min_gpu_memory_bytes = min_gpu_memory_bytes or 0

        self.__pref_cpu_count = None
        self.__pref_memory_bytes = None
        self.__pref_gpu_count = None
        self.__pref_gpu_memory_bytes = None

    # querries

    def groups(self) -> Set[str]:
        return set(self.__groups)

    def min_cpu_count(self) -> Union[float, int]:
        return self.__min_cpu_count

    def min_memory_bytes(self) -> int:
        return self.__min_memory_bytes

    def preferred_cpu_count(self) -> Union[float, int]:
        return self.__pref_cpu_count

    def preferred_memory_bytes(self) -> int:
        return self.__pref_memory_bytes

    def min_gpu_count(self) -> Union[float, int]:
        return self.__min_gpu_count

    def min_gpu_memory_bytes(self) -> int:
        return self.__min_gpu_memory_bytes

    def preferred_gpu_count(self) -> Union[float, int]:
        return self.__pref_gpu_count

    def preferred_gpu_memory_bytes(self) -> int:
        return self.__pref_gpu_memory_bytes

    def worker_type(self) -> WorkerType:
        return self.__worker_type

    # setters

    def set_groups(self, groups):
        self.__groups = groups

    def add_groups(self, groups):
        self.__groups.update(set(x for x in groups if x != ''))

    def add_group(self, group: str):
        self.__groups.add(group)

    def set_min_cpu_count(self, min_cpu_count):
        self.__min_cpu_count = min_cpu_count

    def set_min_memory_bytes(self, min_memory_bytes):
        self.__min_memory_bytes = int(min_memory_bytes)

    def set_preferred_cpu_count(self, pref_cpu_count):
        self.__pref_cpu_count = pref_cpu_count

    def set_preferred_memory_bytes(self, pref_memory_bytes):
        self.__pref_memory_bytes = int(pref_memory_bytes)

    def set_min_gpu_count(self, min_gpu_count):
        self.__min_gpu_count = min_gpu_count

    def set_min_gpu_memory_bytes(self, min_gpu_memory_bytes):
        self.__min_gpu_memory_bytes = int(min_gpu_memory_bytes)

    def set_preferred_gpu_count(self, pref_gpu_count):
        self.__pref_gpu_count = pref_gpu_count

    def set_preferred_gpu_memory_bytes(self, pref_gpu_memory_bytes):
        self.__pref_gpu_memory_bytes = int(pref_gpu_memory_bytes)

    def set_worker_type(self, worker_type: WorkerType):
        self.__worker_type = worker_type

    def final_where_clause(self):
        conds = [f'("worker_type" = {self.__worker_type.value})']
        if self.__min_cpu_count > 0:
            conds.append(f'("cpu_count" >= {self.__min_cpu_count - 1e-8 if isinstance(self.__min_cpu_count, float) else self.__min_cpu_count})')   # to ensure sql compare will work
        if self.__min_memory_bytes > 0:
            conds.append(f'("cpu_mem" >= {self.__min_memory_bytes})')
        if self.__min_gpu_count > 0:
            conds.append(f'("gpu_count" >= {self.__min_gpu_count - 1e-8 if isinstance(self.__min_gpu_count, float) else self.__min_gpu_count})')  # to ensure sql compare will work
        if self.__min_gpu_memory_bytes > 0:
            conds.append(f'("gpu_mem" >= {self.__min_gpu_memory_bytes})')
        if len(self.__groups) > 0:
            esc = '\\'

            def _morph(s: str):
                return re.sub(r'(?<!\\)\?', '_', re.sub(r'(?<!\\)\*', '%', s.replace('%', r'\%').replace('_', r'\_')))
            conds.append(f'''(EXISTS (SELECT * FROM worker_groups wg WHERE wg."worker_hwid" == workers."hwid" AND ( {" OR ".join(f"""wg."group" LIKE '{_morph(x)}' ESCAPE '{esc}'""" for x in self.__groups)} )))''')
        return ' AND '.join(conds)

    def to_dict(self, resources_only: bool = False) -> dict:
        ret = {'cpu_count': self.__min_cpu_count,
               'cpu_mem': self.__min_memory_bytes,
               'gpu_count': self.__min_gpu_count,
               'gpu_mem': self.__min_gpu_memory_bytes}
        if self.__pref_cpu_count is not None:
            ret['pref_cpu_count'] = self.__pref_cpu_count
        if self.__pref_memory_bytes is not None:
            ret['pref_cpu_mem'] = self.__pref_memory_bytes
        if self.__pref_gpu_count is not None:
            ret['pref_gpu_count'] = self.__pref_gpu_count
        if self.__pref_gpu_memory_bytes is not None:
            ret['pref_gpu_mem'] = self.__pref_gpu_memory_bytes

        if resources_only:
            return ret

        ret['groups'] = tuple(self.groups())
        return ret

    def to_min_worker_resources(self) -> WorkerResources:
        res = WorkerResources()
        res.cpu_count = self.__min_cpu_count
        res.cpu_mem = self.__min_memory_bytes
        res.gpu_count = self.__min_gpu_count
        res.gpu_mem = self.__min_gpu_memory_bytes
        return res

    def __gt__(self, other):
        if not isinstance(other, WorkerResources):
            raise NotImplementedError()
        return self.__min_cpu_count > other.cpu_count and \
               self.__min_memory_bytes > other.cpu_mem and \
               self.__min_gpu_count > other.gpu_count and \
               self.__min_gpu_memory_bytes > other.gpu_mem

    def __eq__(self, other):
        if not isinstance(other, WorkerResources):
            raise NotImplementedError()
        return self.__min_cpu_count == other.cpu_count and \
               self.__min_memory_bytes == other.cpu_mem and \
               self.__min_gpu_count == other.gpu_count and \
               self.__min_gpu_memory_bytes == other.gpu_mem


class InvocationJob:
    """
    serializable data about launching something
    """
    def __init__(self, args: List[str], *, env: Optional[InvocationEnvironment] = None, invocation_id=None,
                 requirements: Optional[InvocationRequirements] = None,
                 # environment_wrapper_arguments: Optional[EnvironmentResolverArguments] = None,
                 good_exitcodes: Optional[Iterable[int]] = None,
                 retry_exitcodes: Optional[Iterable[int]] = None):
        """

        :param args: list of args passed to exec
        :param env: extra special environment variables to be set. in most cases you don't need that
        :param invocation_id: invocation ID in scheduler's DB this invocation is connected to
        :param requirements: requirements for the worker to satisfy in order to be able to pick up this invocation job
        # :param environment_wrapper_arguments: environment wrapper and arguments to pass to it. if None - worker's configured default wrapper will be invoked
        :param good_exitcodes: set of process exit codes to consider as success. default would be just 0
        :param retry_exitcodes: set of process exit codes to consider as retry is needed. for ex in case of sidefx products exit code 3 means there was a license error
                                which can occur due to delay in license return or network problems or bad license management.
                                in that case you might just want to restart the invocation job automatically. there are no codes in this set by default.
        """
        self.__args = [str(arg) for arg in args]
        self.__env = env or InvocationEnvironment()
        self.__invocation_id = invocation_id
        self.__task_id = None

        self.__out_progress_regex = re.compile(rb'ALF_PROGRESS\s+(\d+)%')
        self.__err_progress_regex = None

        self.__requirements = requirements or InvocationRequirements()
        self.__priority = 0.0
        self.__envres_args = None  # environment_wrapper_arguments

        self.__exitcode = None
        self.__running_time = None
        self.__good_exitcodes = set(good_exitcodes or [0])
        self.__retry_exitcodes = set(retry_exitcodes or [])

        self.__attrs = {}
        self.__extra_files: Dict[str, Union[str, bytes]] = {}

    def requirements(self) -> InvocationRequirements:
        return self.__requirements

    def set_requirements(self, requirements: InvocationRequirements):
        self.__requirements = requirements

    def priority(self) -> float:
        return self.__priority

    def set_priority(self, priority: float):
        self.__priority = priority

    def environment_resolver_arguments(self):  # type: () -> Optional[EnvironmentResolverArguments]
        return self.__envres_args

    def set_stdout_progress_regex(self, regex: Optional[bytes]):
        """
        regex logic is the following:
            - if it has one group - that is treated as progress percentage
            - if it has 2 or more unnamed groups - those are treated as <group1> out of <group2> progress
              so output = g1/g2*100 %
            - if there are named groups named 'current' and 'total' - those groups are treated the same way
              as above: = current/total*100 %
        """
        if regex is None:
            self.__out_progress_regex = None
            return
        self.__out_progress_regex = re.compile(regex)

    def set_stderr_progress_regex(self, regex: Optional[bytes]):
        """
        regex logic is the following:
            - if it has one group - that is treated as progress percentage
            - if it has 2 or more unnamed groups - those are treated as <group1> out of <group2> progress
              so output = g1/g2*100 %
            - if there are named groups named 'current' and 'total' - those groups are treated the same way
              as above: = current/total*100 %
        """
        if regex is None:
            self.__err_progress_regex = None
            return
        self.__err_progress_regex = re.compile(regex)

    def set_extra_file(self, file_path: str, file_data: Union[str, bytes]):
        """
        extra files to transfer with this task
        these files can be referenced in args list in qt resources format:
        like:
        :/path/to/file.smth
        if such file does not exist - argument will be left as is

        WARNING:  slashes / are used to split dirs, OS/FS INDEPENDENT
        WARNING:  other than that it is YOUR responsibility to ensure file names do not contain illegal characters

        :param file_path: file path, like file.smth or path/to/file.smth NOTE: NO leading slashes
        :param file_data: data. str data will be saved as utf-8 text, binary data will be saved as is
        """
        self.__extra_files[file_path] = file_data

    def match_stdout_progress(self, line: bytes) -> Optional[float]:
        if self.__out_progress_regex is None:
            return None
        try:
            return self._match_progress(line, self.__out_progress_regex)
        except BadProgressRegexp:
            self.__out_progress_regex = None
            return None

    def match_stderr_progress(self, line: bytes) -> Optional[float]:
        if self.__err_progress_regex is None:
            return None
        try:
            return self._match_progress(line, self.__err_progress_regex)
        except BadProgressRegexp:
            self.__err_progress_regex = None
            return None

    @classmethod
    def _match_progress(cls, line: bytes, regex: re.Pattern) -> Optional[float]:
        """
        should be given a bytes line as produced by running task output
        will return progress percentage (so 0-100) float, if output line matches progress regex
        otherwise returns None

        regex logic is the following:
            - if it has one group - that is treated as progress percentage
            - if it has 2 or more unnamed groups - those are treated as <group1> out of <group2> progress
              so output = g1/g2*100 %
            - if there are named groups named 'current' and 'total' - those groups are treated the same way
              as above: = current/total*100 %
        """
        match = regex.match(line)
        if match is None:
            return
        if len(match.groups()) == 0:
            raise BadProgressRegexp()
        # several cases possible
        if len(match.groups()) == 1:  # just single group match - treated as progress percentage
            return float(match.group(1))
        named_groups = match.groupdict()
        if len(named_groups) == 0:  # no named groups - first 2 groups are treated as <g1> out of <g2>
            return float(match.group(1)) / float(match.group(2)) * 100.0
        if 'current' in named_groups and 'total' in named_groups:
            return float(match.group('current')) / float(match.group('total')) * 100.0

    def attributes(self):
        return MappingProxyType(self.__attrs)

    def extra_files(self):
        return MappingProxyType(self.__extra_files)

    def args(self) -> List[str]:
        return self.__args

    def env(self):
        return self.__env

    def invocation_id(self) -> int:
        return self.__invocation_id

    def task_id(self) -> int:
        return self.__task_id

    def finish(self, exitcode: int, running_time: float):
        self.__exitcode = exitcode
        self.__running_time = running_time

    def is_finished(self):
        return self.__exitcode is not None

    def exit_code(self):
        return self.__exitcode

    def running_time(self) -> Optional[float]:
        """
        return time of the invocation run time,
        or None if invocation was not finished
        """
        return self.__running_time

    def finished_with_error(self):
        if self.__exitcode is None:
            raise InvocationNotFinished()
        return self.__exitcode not in self.__good_exitcodes

    def finished_needs_retry(self):
        if self.__exitcode is None:
            raise InvocationNotFinished()
        return self.__exitcode in self.__retry_exitcodes

    async def serialize_async(self) -> bytes:
        return await asyncio.get_event_loop().run_in_executor(None, pickle.dumps, self)

    #
    # methods for scheduler
    def _set_invocation_id(self, invocation_id):
        self.__invocation_id = invocation_id

    def _set_task_id(self, task_id):
        self.__task_id = task_id

    def _set_task_attributes(self, attr_dict):
        self.__attrs = deepcopy(attr_dict)

    def _set_envresolver_arguments(self, args: "EnvironmentResolverArguments"):
        self.__envres_args = args

    def __repr__(self):
        return f'InvocationJob: {self.__invocation_id}, {repr(self.__args)} {repr(self.__env)}'

    @classmethod
    def deserialize(cls, data: bytes) -> "InvocationJob":
        return pickle.loads(data)

    @classmethod
    async def deserialize_async(cls, data: bytes) -> "InvocationJob":
        return await asyncio.get_event_loop().run_in_executor(None, cls.deserialize, data)



import json
from .clients import JsonMessageClient, CommandJsonMessageClient, JsonMessageWrapper
from ..client import MessageClient
from ..message_handler import MessageHandlerBase
from ..messages import Message
from ..logging import get_logger

from typing import Awaitable, Callable, Dict


class JsonMessageHandlerBase(MessageHandlerBase):
    def __init__(self):
        super().__init__()
        self._logger = get_logger(self.__class__.__name__)

    async def process_message(self, message: Message, client: MessageClient) -> bool:
        if not isinstance(client, JsonMessageClient):
            # we expect client to be json client
            return False
        jmessage = JsonMessageWrapper(message)
        try:
            await jmessage.message_body_as_json()
        except (json.JSONDecodeError, UnicodeDecodeError):
            self._logger.exception('failed to parse json')
            return False
        return await self.process_json_message(jmessage, client) or False

    async def process_json_message(self, message: JsonMessageWrapper, client: JsonMessageClient) -> bool:
        raise NotImplementedError()


class CommandMessageHandlerBase(JsonMessageHandlerBase):
    def command_mapping(self) -> Dict[str, Callable[[dict, CommandJsonMessageClient, Message], Awaitable[None]]]:
        raise NotImplementedError()

    async def process_json_message(self, message: JsonMessageWrapper, client: CommandJsonMessageClient) -> bool:
        if not isinstance(client, CommandJsonMessageClient):
            return False

        body = await message.message_body_as_json()
        if 'command' in body:
            command_json = body['command']
            command = command_json['name']
            args = command_json.get('arguments', {})

            commands_map = self.command_mapping()
            if command not in commands_map:
                return False

            await commands_map[command](args, client, message)
            return True
        else:
            self._logger.error(f'message is not a command, "{message}"')
            return False



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionDepthMap


class AlicevisionDepthMap(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionDepthMap, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Structure From Motion', NodeParameterType.STRING, "`task['av_structure_from_motion']`")
            ui.add_parameter('imagesFolder', 'Dense Images', NodeParameterType.STRING, "`task['av_prepare_dense']`")
            ui.add_parameter('output', 'Output Depth Map Folder', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/DepthMap")
            ui.add_separator()
            ui.add_parameter('downscale', 'Downscale', NodeParameterType.INT, 2)\
                .add_menu((('1', 1), ('2', 2), ('3', 3), ('4', 4)))
            ui.add_parameter('sgmMaxTCams', 'SGM: Nb Neighbour Cameras', NodeParameterType.INT, 10)
            ui.add_parameter('refineMaxTCams', 'Refine: Nb Neighbour Cameras', NodeParameterType.INT, 6)

            ui.add_separator()
            with ui.parameters_on_same_line_block():
                ui.add_parameter('rangeStart', 'Image Range Start', NodeParameterType.INT, 0)
                ui.add_parameter('rangeSize', 'Image Chunk Size', NodeParameterType.INT, 999999)

    @classmethod
    def label(cls) -> str:
        return 'AV Depth Map'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'depth', 'map'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_depth_map'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('output').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.mkdir(parents=True, exist_ok=True)

        args = ['aliceVision_depthMapEstimation', '--verboseLevel', 'info']
        args += ['--input', context.param_value('input').strip()]
        args += ['--imagesFolder', context.param_value('imagesFolder').strip()]
        args += ['--sgmMaxTCams', context.param_value('sgmMaxTCams')]
        args += ['--refineMaxTCams', context.param_value('refineMaxTCams')]
        for param_name in ('rangeStart', 'rangeSize'):
            args += [f'--{param_name}', context.param_value(param_name)]
        args += ['--output', output]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_depth_map', str(output))
        return res



import asyncio
import json
import pickle
import dataclasses
from dataclasses import dataclass

from .base import TypeMetadata
from .enums import NodeParameterType

from typing import Any, Optional, Tuple, Dict, Set, Iterable


class NodeSnippetData(TypeMetadata):
    """
    class containing enough information to reproduce a certain snippet of nodes, with parameter values and connections ofc
    """

    class Serializer(json.JSONEncoder):
        def default(self, obj):
            if dataclasses.is_dataclass(obj):
                dcs = obj.__dict__  # dataclasses.asdict is recursive, kills inner dataclasses
                dcs['__dataclass__'] = obj.__class__.__name__
                return dcs
            elif isinstance(obj, NodeSnippetData):
                return {'nodes': obj.nodes_data,
                        'connections': obj.connections_data,
                        'label': obj.label,
                        'tags': list(obj.tags),
                        '__NodeSnippetData__': '==3*E==',
                        '__format_version__': [1, 0, 0]
                        }
            elif isinstance(obj, NodeParameterType):
                return {'value': obj.value,
                        '__NodeParameterType__': '==3*E=='
                        }
            return super(NodeSnippetData.Serializer, self).default(obj)

    class Deserializer(json.JSONDecoder):
        def dedata(self, obj):
            if '__dataclass__' in obj:
                data = getattr(NodeSnippetData, obj['__dataclass__'])(**{k: v for k, v in obj.items() if k != '__dataclass__'})
                if obj['__dataclass__'] == 'NodeData':
                    data.pos = tuple(data.pos)
                return data
            elif obj.get('__NodeSnippetData__', None) == '==3*E==':
                if obj['__format_version__'] >= [2]:
                    raise NotImplementedError(f'snippet format {obj["__format_version__"]} is not supported')
                return NodeSnippetData(obj['nodes'], obj['connections'], obj['label'], obj['tags'])
            elif obj.get('__NodeParameterType__', None) == '==3*E==':
                return NodeParameterType(obj['value'])
            return obj

        def __init__(self):
            super(NodeSnippetData.Deserializer, self).__init__(object_hook=self.dedata)

    @dataclass
    class ParamData:
        name: str
        type: NodeParameterType
        uvalue: Any
        expr: Optional[str]

    @dataclass
    class NodeData:
        tmpid: int
        type: str
        name: str
        parameters: Dict[str, "NodeSnippetData.ParamData"]
        pos: Tuple[float, float]

    @dataclass
    class ConnData:
        tmpout: int
        out_name: str
        tmpin: int
        in_name: str

    @property
    def nodes_data(self):
        return self.__nodes_data

    @nodes_data.setter
    def nodes_data(self, nodes_data):
        self.__nodes_data = nodes_data
        self.__avgpos = None

    @property
    def connections_data(self):
        return self.__connections_data

    @connections_data.setter
    def connections_data(self, connections_data):
        self.__connections_data = connections_data

    @property
    def pos(self) -> Tuple[float, float]:
        if self.__avgpos is None:
            avgpos = [0., 0.]
            for nodedata in self.__nodes_data:
                avgpos[0] += nodedata.pos[0]
                avgpos[1] += nodedata.pos[1]
            nnodes = max(1, len(self.__nodes_data))
            self.__avgpos = (avgpos[0] / nnodes, avgpos[1] / nnodes)
        return self.__avgpos

    @property
    def label(self) -> Optional[str]:
        return self.__label

    @property
    def tags(self) -> Set[str]:
        return self.__tags

    def add_tag(self, tag: str):
        self.__tags.add(tag)

    def remove_node_temp_ids_from_snippet(self, node_temp_ids: Iterable[int]):
        to_rm = set(node_temp_ids)
        self.__nodes_data = [nd for nd in self.__nodes_data if nd.tmpid not in to_rm]
        in_snippet = set(nd.tmpid for nd in self.__nodes_data)
        self.__connections_data = [cd for cd in self.__connections_data
                                   if cd.tmpout in in_snippet or cd.tmpin in in_snippet]

    def __init__(self, nodes_data: Iterable[NodeData], connections_data: Iterable[ConnData], label: Optional[str] = None, tags: Optional[Iterable[str]] = None):
        self.__nodes_data = list(nodes_data)
        self.__connections_data = list(connections_data)
        self.__label: Optional[str] = label
        self.__tags: Set[str] = set(tags) if tags is not None else set()
        self.__avgpos = None

    def __eq__(self, other: "NodeSnippetData"):
        return self.nodes_data == other.nodes_data and self.connections_data == other.connections_data and \
               self.__tags == other.__tags and \
               self.__label == other.__label

    def __repr__(self):
        nodesrepr = '\n'.join(repr(x) for x in self.nodes_data)
        connsrepr = '\n'.join(repr(x) for x in self.connections_data)
        return f'NodeSnippetData: nodes: {nodesrepr}\n\nconnections: {connsrepr}'

    def serialize(self, ascii=True) -> bytes:
        """
        serialize into bytes, ascii-friendly or not
        """
        if ascii:
            return json.dumps(self, cls=NodeSnippetData.Serializer, indent=4).encode('UTF-8')
        return b'\0' + pickle.dumps(self)

    async def serialize_async(self) -> bytes:
        return await asyncio.get_event_loop().run_in_executor(None, self.serialize)

    @classmethod
    def deserialize(cls, data: bytes) -> "NodeSnippetData":
        if data[0] == 0:
            return pickle.loads(data[1:])
        else:
            return json.loads(data.decode('UTF-8'), cls=NodeSnippetData.Deserializer)

    @classmethod
    async def deserialize_async(cls, data: bytes) -> "NodeSnippetData":
        return await asyncio.get_event_loop().run_in_executor(None, cls.deserialize, data)


class NodeSnippetDataPlaceholder(TypeMetadata):
    """
    this is an empty metadata-only class to transfer over the net
    """
    def __init__(self, label: Optional[str] = None, tags: Optional[Iterable[str]] = None):
        self.__label = label
        self.__tags = set(tags) if tags is not None else set()

    @property
    def label(self) -> Optional[str]:
        return self.__label

    @property
    def tags(self) -> Set[str]:
        return self.__tags

    @classmethod
    def from_nodesnippetdata(cls, source: NodeSnippetData):
        return NodeSnippetDataPlaceholder(source.label, source.tags)



from copy import copy
from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment

from typing import Iterable


def node_class():
    return Redshift


class Redshift(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'redshift'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'houdini', 'redshift', 'rs', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'redshift'

    def __init__(self, name):
        super(Redshift, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_parameter('rs path', 'rs file path', NodeParameterType.STRING, "`task['file']`")
            ui.add_parameter('image path', 'output image file path', NodeParameterType.STRING, "`task['outimage']`")
            ui.add_parameter('skip if exists', 'skip if result already exists', NodeParameterType.BOOL, False)

            # ui.parameter('')
            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    def process_task(self, context) -> ProcessingResult:
        args = context.task_attributes()

        env = InvocationEnvironment()

        # TODO!!!  in case of skip existing "files" is not filled with AOVs !!!

        script = 'import os\n' \
                 'import sys\n' \
                 'from subprocess import Popen, PIPE\n' \
                 'import tempfile\n' \
                 'import shutil\n' \
                 'import lifeblood_connection as lbc\n' \
                 '\n' \
                 'out_beauty = {out_beauty}\n' \
                 'if {skip_if_exists} and os.path.exists(out_beauty):\n' \
                 "    print('image file already exists, skipping work')\n" \
                 '    lbc.set_attributes({{"file": out_beauty, \n' \
                 '                         "files": [out_beauty]}}, blocking=True)\n' \
                 "    sys.exit(0)\n" \
                 "\n" \
                 'temp_render_dir = tempfile.mkdtemp(prefix="redshift_")\n' \
                 "output_files = []\n" \
                 "p = Popen(['redshiftCmdLine', {rspath}, '-oip', temp_render_dir], stdout=PIPE)\n" \
                 'while p.poll() is None:\n' \
                 '    full_line = False\n' \
                 '    part_parts = []\n' \
                 '    while not full_line:\n' \
                 '        part = p.stdout.readline(1024)\n' \
                 '        part_parts.append(part)\n' \
                 '        if part.endswith(b"\\n") or part == b"" or p.poll() is not None:\n' \
                 '            full_line = True\n' \
                 '            break\n' \
                 '    part = b"".join(part_parts)\n' \
                 '\n' \
                 '    if part.startswith(b"Saving: "):\n' \
                 '        output_files.append(part[8:].strip())\n' \
                 '    sys.stdout.buffer.write(part)  # "promote" to stdout\n' \
                 '\n' \
                 'if not output_files:\n' \
                 '    sys.exit(1)\n' \
                 'output_files = [x.decode("utf-8") for x in output_files]\n' \
                 '\n' \
                 'def cleancopy(src, dst):\n' \
                 '    try:\n' \
                 '        os.unlink(dst)\n' \
                 '    except FileNotFoundError:\n' \
                 '        pass\n' \
                 '    os.makedirs(os.path.dirname(dst), exist_ok=True)\n' \
                 '    shutil.copy2(src, dst)\n' \
                 '\n' \
                 'print("copying locally rendered files to final destination...")\n' \
                 'final_output_files = []\n' \
                 'sys.stdout.buffer.write(f"copying to {{out_beauty}}\\n".encode("UTF-8"))\n' \
                 'cleancopy(output_files[0], out_beauty)\n' \
                 'final_output_files.append(out_beauty)\n' \
                 'out_beauty_dir = os.path.dirname(out_beauty)\n' \
                 'for file_path in output_files:\n' \
                 '    file_path_dst = os.path.join(out_beauty_dir, os.path.basename(file_path))\n' \
                 '    sys.stdout.buffer.write(f"copying to {{file_path_dst}}\\n".encode("UTF-8"))\n' \
                 '    cleancopy(file_path, file_path_dst)\n' \
                 '    final_output_files.append(file_path_dst)\n' \
                 '\n' \
                 'lbc.set_attributes({{"file": final_output_files[0],\n' \
                 '                     "files": final_output_files}}, blocking=True)\n' \
                 'print("cleaning up...")\n' \
                 'shutil.rmtree(temp_render_dir)\n' \
                 'print("all done")\n' \
                 "sys.exit()\n" \
                 '' \
                 .format(skip_if_exists=repr(context.param_value('skip if exists')),  # for now we cannot know output image until after render
                         rspath=repr(context.param_value('rs path')),
                         out_beauty=repr(context.param_value('image path')))

        invoc = InvocationJob(['python', ':/rsccall.py'])
        invoc.set_extra_file('rsccall.py', script)
        invoc.set_stdout_progress_regex(rb'^\s*Block\s+(\d+)/(\d+)\s+')

        res = ProcessingResult(invoc)
        return res



import io

import lz4.frame
import struct
from io import BytesIO, BufferedIOBase
from .buffered_connection import BufferedReader
from .buffer_serializable import IBufferSerializable
from .enums import TaskState, WorkerState, WorkerType, TaskGroupArchivedState, InvocationState
from dataclasses import dataclass

from typing import Dict, List, Tuple, Type, Optional, Set, Union


def _serialize_string(s: str, stream: BufferedIOBase) -> int:
    bstr = s.encode('UTF-8')
    stream.write(struct.pack('>Q', len(bstr)))
    return stream.write(bstr)


def _deserialize_string(stream: BufferedReader) -> str:
    bsize, = struct.unpack('>Q', stream.readexactly(8))
    if bsize == 0:
        return ''
    return bytes(stream.readexactly(bsize)).decode('UTF-8')


class DataNotSet:
    pass


@dataclass
class TaskDelta(IBufferSerializable):
    id: int
    parent_id: Union[Optional[int], Type[DataNotSet]] = DataNotSet
    children_count: Union[int, Type[DataNotSet]] = DataNotSet                   # TODO: remove, add to getting individual task details
    active_children_count: Union[int, Type[DataNotSet]] = DataNotSet            # TODO: remove, add to getting individual task details
    state: Union[TaskState, Type[DataNotSet]] = DataNotSet                      # TODO: remove, add to getting individual task details
    state_details: Union[Optional[str], Type[DataNotSet]] = DataNotSet          # TODO: remove, add to getting individual task details
    paused: Union[bool, Type[DataNotSet]] = DataNotSet
    node_id: Union[int, Type[DataNotSet]] = DataNotSet
    node_input_name: Union[Optional[str], Type[DataNotSet]] = DataNotSet
    node_output_name: Union[Optional[str], Type[DataNotSet]] = DataNotSet
    name: Union[str, Type[DataNotSet]] = DataNotSet
    split_level: Union[int, Type[DataNotSet]] = DataNotSet                      # TODO: is this needed for UI?
    work_data_invocation_attempt: Union[int, Type[DataNotSet]] = DataNotSet     # TODO: remove, add to getting individual task details
    progress: Union[Optional[float], Type[DataNotSet]] = DataNotSet
    split_origin_task_id: Union[Optional[int], Type[DataNotSet]] = DataNotSet   # used by UI to draw new tasks from the split_origin_task_id task
    split_id: Union[Optional[int], Type[DataNotSet]] = DataNotSet               # TODO: is this needed for UI?
    invocation_id: Union[Optional[int], Type[DataNotSet]] = DataNotSet
    groups: Union[Set[str], Type[DataNotSet]] = DataNotSet

    def serialize(self, stream: BufferedIOBase):
        data_has = struct.pack(
            '>?????????????????',
            *(x is not DataNotSet for x in
              (self.parent_id, self.children_count, self.active_children_count, self.state, self.state_details,
               self.paused, self.node_id, self.node_input_name, self.node_output_name, self.name, self.split_level,
               self.work_data_invocation_attempt, self.progress, self.split_origin_task_id, self.split_id,
               self.invocation_id, self.groups)))
        stream.write(data_has)

        pattern_parts = ['>Q']
        pack_stuff = [self.id]
        if self.parent_id is not DataNotSet:
            pattern_parts.append('?Q')
            pack_stuff.extend([self.parent_id is not None, self.parent_id or 0])
        if self.children_count is not DataNotSet:
            pattern_parts.append('Q')
            pack_stuff.append(self.children_count)
        if self.active_children_count is not DataNotSet:
            pattern_parts.append('Q')
            pack_stuff.append(self.active_children_count)
        if self.state is not DataNotSet:
            pattern_parts.append('I')
            pack_stuff.append(self.state.value)
        if self.paused is not DataNotSet:
            pattern_parts.append('?')
            pack_stuff.append(self.paused)
        if self.node_id is not DataNotSet:
            pattern_parts.append('Q')
            pack_stuff.append(self.node_id)
        if self.split_level is not DataNotSet:
            pattern_parts.append('Q')
            pack_stuff.append(self.split_level)
        if self.work_data_invocation_attempt is not DataNotSet:
            pattern_parts.append('Q')
            pack_stuff.append(self.work_data_invocation_attempt)
        if self.progress is not DataNotSet:
            pattern_parts.append('?d')
            pack_stuff.extend([self.progress is not None, self.progress or 0.0])
        if self.split_origin_task_id is not DataNotSet:
            pattern_parts.append('?Q')
            pack_stuff.extend([self.split_origin_task_id is not None, self.split_origin_task_id or 0])
        if self.split_id is not DataNotSet:
            pattern_parts.append('?Q')
            pack_stuff.extend([self.split_id is not None, self.split_id or 0])
        if self.invocation_id is not DataNotSet:
            pattern_parts.append('?Q')
            pack_stuff.extend([self.invocation_id is not None, self.invocation_id or 0])
        if self.groups is not DataNotSet:
            pattern_parts.append('Q')
            pack_stuff.append(len(self.groups))

        data = struct.pack(''.join(pattern_parts), *pack_stuff)
        stream.write(data)

        if self.state_details is not DataNotSet:
            stream.write(struct.pack('>?', self.state_details is not None))
            if self.state_details is not None:
                _serialize_string(self.state_details, stream)
        if self.node_input_name is not DataNotSet:
            stream.write(struct.pack('>?', self.node_input_name is not None))
            if self.node_input_name is not None:
                _serialize_string(self.node_input_name, stream)
        if self.node_output_name is not DataNotSet:
            stream.write(struct.pack('>?', self.node_output_name is not None))
            if self.node_output_name is not None:
                _serialize_string(self.node_output_name, stream)
        if self.name is not DataNotSet:
            _serialize_string(self.name, stream)
        if self.groups is not DataNotSet:
            for group in self.groups:
                _serialize_string(group, stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader):
        has_parent_id, has_children_count, \
            has_active_children_count, has_state, has_state_details, has_paused, has_node_id, has_node_input_name, \
            has_node_output_name, has_name, has_split_level, has_work_data_invocation_attempt, \
            has_progress, has_split_origin_task_id, has_split_id, has_invocation_id, has_groups = \
                struct.unpack('>?????????????????', stream.readexactly(17))

        task_id, = struct.unpack('>Q', stream.readexactly(8))
        parent_id = DataNotSet
        if has_parent_id:
            is_not_none, parent_id = struct.unpack('>?Q', stream.readexactly(9))
            if not is_not_none:
                parent_id = None
        children_count = DataNotSet
        if has_children_count:
            children_count, = struct.unpack('>Q', stream.readexactly(8))
        active_children_count = DataNotSet
        if has_active_children_count:
            active_children_count, = struct.unpack('>Q', stream.readexactly(8))
        state = DataNotSet
        if has_state:
            _state_raw, = struct.unpack('>I', stream.readexactly(4))
            state = TaskState(_state_raw)
        paused = DataNotSet
        if has_paused:
            paused, = struct.unpack('>?', stream.readexactly(1))
        node_id = DataNotSet
        if has_node_id:
            node_id, = struct.unpack('>Q', stream.readexactly(8))
        split_level = DataNotSet
        if has_split_level:
            split_level, = struct.unpack('>Q', stream.readexactly(8))
        work_data_invocation_attempt = DataNotSet
        if has_work_data_invocation_attempt:
            work_data_invocation_attempt, = struct.unpack('>Q', stream.readexactly(8))
        progress = DataNotSet
        if has_progress:
            is_not_none, progress = struct.unpack('>?d', stream.readexactly(9))
            if not is_not_none:
                progress = None
        split_origin_task_id = DataNotSet
        if has_split_origin_task_id:
            is_not_none, split_origin_task_id = struct.unpack('>?Q', stream.readexactly(9))
            if not is_not_none:
                split_origin_task_id = None
        split_id = DataNotSet
        if has_split_id:
            is_not_none, split_id = struct.unpack('>?Q', stream.readexactly(9))
            if not is_not_none:
                split_id = None
        invocation_id = DataNotSet
        if has_invocation_id:
            is_not_none, invocation_id = struct.unpack('>?Q', stream.readexactly(9))
            if not is_not_none:
                invocation_id = None
        group_count = 0
        if has_groups:
            group_count, = struct.unpack('>Q', stream.readexactly(8))

        state_details = DataNotSet
        if has_state_details:
            state_details = None
            is_not_none, = struct.unpack('>?', stream.readexactly(1))
            if is_not_none:
                state_details = _deserialize_string(stream)

        node_input_name = DataNotSet
        if has_node_input_name:
            node_input_name = None
            has_node_input_name, = struct.unpack('>?', stream.readexactly(1))
            if has_node_input_name:
                node_input_name = _deserialize_string(stream)
        node_output_name = DataNotSet
        if has_node_output_name:
            node_output_name = None
            has_node_output_name, = struct.unpack('>?', stream.readexactly(1))
            if has_node_output_name:
                node_output_name = _deserialize_string(stream)
        task_name = DataNotSet
        if has_name:
            task_name = _deserialize_string(stream)
        groups = DataNotSet
        if has_groups:
            groups = set()
            for i in range(group_count):
                group_name = _deserialize_string(stream)
                groups.add(group_name)

        return TaskDelta(task_id, parent_id, children_count, active_children_count, state, state_details, paused,
                         node_id, node_input_name, node_output_name, task_name, split_level, work_data_invocation_attempt,
                         progress, split_origin_task_id, split_id, invocation_id, groups)

    def __repr__(self):
        parts = []
        for field, val in self.__dict__.items():
            if val is DataNotSet:
                continue
            parts.append(f'{field}={val}')
        return f'TaskDelta({", ".join(parts)})'

    def tiny_repr(self):
        return f'{type(self).__name__}:[{self.id}]'


@dataclass
class TaskData(IBufferSerializable):
    id: int
    parent_id: Optional[int]
    children_count: int
    active_children_count: int
    state: TaskState
    state_details: Optional[str]
    paused: bool
    node_id: int
    node_input_name: Optional[str]
    node_output_name: Optional[str]
    name: str
    split_level: int
    work_data_invocation_attempt: int
    progress: Optional[float]
    split_origin_task_id: Optional[int]
    split_id: Optional[int]
    invocation_id: Optional[int]
    groups: Set[str]

    def serialize(self, stream: BufferedIOBase):
        #                    i?pcaspnsw?p?s?s?ig
        data = struct.pack('>Q?QQQI?QQQ?d?Q?Q?QQ', self.id, self.parent_id is not None, self.parent_id or 0, self.children_count,
                           self.active_children_count, self.state.value, self.paused, self.node_id, self.split_level,
                           self.work_data_invocation_attempt, self.progress is not None, self.progress or 0.0,
                           self.split_origin_task_id is not None, self.split_origin_task_id or 0,
                           self.split_id is not None, self.split_id or 0,
                           self.invocation_id is not None, self.invocation_id or 0, len(self.groups))
        stream.write(data)
        stream.write(struct.pack('>?', self.state_details is not None))
        if self.state_details is not None:
            _serialize_string(self.state_details, stream)
        stream.write(struct.pack('>?', self.node_input_name is not None))
        if self.node_input_name is not None:
            _serialize_string(self.node_input_name, stream)
        stream.write(struct.pack('>?', self.node_output_name is not None))
        if self.node_output_name is not None:
            _serialize_string(self.node_output_name, stream)
        _serialize_string(self.name, stream)
        for group in self.groups:
            _serialize_string(group, stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "TaskData":
        offset = 106  # 8 + 1 + 8*3 + 4 + 1 + 8*3 + 1+8 + 1+8 + 1+8 + 1+8 + 8
        task_id, has_parent_id, task_parent_id, task_children_count, task_active_children_count, \
            task_state_value, task_paused, task_node_id, task_split_level, task_work_data_invocation_attempt, \
            has_progress, progress, \
            has_task_split_origin_task_id, task_split_origin_task_id, \
            has_task_split_id, task_split_id, \
            has_task_invocation_id, task_invocation_id, group_count = struct.unpack('>Q?QQQI?QQQ?d?Q?Q?QQ', stream.readexactly(offset))

        has_state_details, = struct.unpack('>?', stream.readexactly(1))
        if has_state_details:
            state_details = _deserialize_string(stream)
        else:
            state_details = None
        has_node_input_name, = struct.unpack('>?', stream.readexactly(1))
        if has_node_input_name:
            node_input_name = _deserialize_string(stream)
        else:
            node_input_name = None
        has_node_output_name, = struct.unpack('>?', stream.readexactly(1))
        if has_node_output_name:
            node_output_name = _deserialize_string(stream)
        else:
            node_output_name = None
        task_name = _deserialize_string(stream)

        groups = set()
        for i in range(group_count):
            group_name = _deserialize_string(stream)
            groups.add(group_name)

        return TaskData(task_id, task_parent_id if has_parent_id else None, task_children_count, task_active_children_count,
                        TaskState(task_state_value), state_details, task_paused, task_node_id, node_input_name, node_output_name,
                        task_name, task_split_level, task_work_data_invocation_attempt,
                        progress if has_progress else None,
                        task_split_origin_task_id if has_task_split_origin_task_id else None,
                        task_split_id if has_task_split_id else None,
                        task_invocation_id if has_task_invocation_id else None, groups)

    def tiny_repr(self):
        return f'{type(self).__name__}:[{self.id}]'


@dataclass
class TaskBatchData(IBufferSerializable):
    db_uid: int
    tasks: Dict[int, TaskData]

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQ', self.db_uid, len(self.tasks)))
        for task in self.tasks.values():
            task.serialize(stream)

    @classmethod
    def deserialize(cls, stream: BufferedIOBase) -> "TaskBatchData":
        db_uid, tasks_count, = struct.unpack('>QQ', stream.readexactly(16))
        tasks = {}
        for i in range(tasks_count):
            task_data = TaskData.deserialize(stream)
            tasks[task_data.id] = task_data
        return TaskBatchData(db_uid, tasks)

    def tiny_repr(self):
        return f'{type(self).__name__}:[{",".join(str(x) for x in self.tasks.keys())}]'


@dataclass
class NodeData(IBufferSerializable):
    id: int
    name: str
    type: str

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>Q', self.id))
        _serialize_string(self.name, stream)
        _serialize_string(self.type, stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "NodeData":
        node_id, = struct.unpack('>Q', stream.readexactly(8))
        node_name = _deserialize_string(stream)
        node_type = _deserialize_string(stream)
        return NodeData(node_id, node_name, node_type)


@dataclass
class NodeConnectionData(IBufferSerializable):
    connection_id: int
    in_id: int
    in_name: str
    out_id: int
    out_name: str

    def serialize(self, stream: BufferedIOBase):
        chunk = struct.pack('>QQQ', self.connection_id, self.in_id, self.out_id)
        stream.write(chunk)
        _serialize_string(self.in_name, stream)
        _serialize_string(self.out_name, stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "NodeConnectionData":
        connection_id, in_id, out_id = struct.unpack('>QQQ', stream.readexactly(24))
        in_name = _deserialize_string(stream)
        out_name = _deserialize_string(stream)
        return NodeConnectionData(connection_id, in_id, in_name, out_id, out_name)


@dataclass
class NodeGraphStructureData(IBufferSerializable):
    db_uid: int
    nodes: Dict[int, NodeData]
    connections: Dict[int, NodeConnectionData]

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQQ', self.db_uid, len(self.nodes), len(self.connections)))
        for node in self.nodes.values():
            node.serialize(stream)
        for connection in self.connections.values():
            connection.serialize(stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "NodeGraphStructureData":
        db_uid, nodes_count, connections_count = struct.unpack('>QQQ', stream.readexactly(24))
        nodes = {}
        connections = {}
        for i in range(nodes_count):
            node_data = NodeData.deserialize(stream)
            nodes[node_data.id] = node_data
        for i in range(connections_count):
            conn_data = NodeConnectionData.deserialize(stream)
            connections[conn_data.connection_id] = conn_data
        return NodeGraphStructureData(db_uid, nodes, connections)


@dataclass
class WorkerResources(IBufferSerializable):
    cpu_count: float
    total_cpu_count: float
    cpu_mem: int
    total_cpu_mem: int
    gpu_count: float
    total_gpu_count: float
    gpu_mem: int
    total_gpu_mem: int

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>ddQQddQQ', self.cpu_count, self.total_cpu_count, self.cpu_mem, self.total_cpu_mem,
                                 self.gpu_count, self.total_gpu_count, self.gpu_mem, self.total_gpu_mem))

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "WorkerResources":
        cpu_count, total_cpu_count, cpu_mem, total_cpu_mem, \
        gpu_count, total_gpu_count, gpu_mem, total_gpu_mem = struct.unpack('>ddQQddQQ', stream.readexactly(64))
        return WorkerResources(cpu_count, total_cpu_count, cpu_mem, total_cpu_mem,
                               gpu_count, total_gpu_count, gpu_mem, total_gpu_mem)


@dataclass
class WorkerData(IBufferSerializable):
    id: int
    worker_resources: WorkerResources
    hwid: str
    last_address: str
    last_seen_timestamp: int
    state: WorkerState
    type: WorkerType
    current_invocation_node_id: Optional[int]
    current_invocation_task_id: Optional[int]
    current_invocation_id: Optional[int]
    current_invocation_progress: Optional[float]
    groups: Set[str]

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQII?Q?Q?Q?dQ', self.id, self.last_seen_timestamp, self.state.value, self.type.value,
                                 self.current_invocation_node_id is not None, self.current_invocation_node_id or 0,
                                 self.current_invocation_task_id is not None, self.current_invocation_task_id or 0,
                                 self.current_invocation_id is not None, self.current_invocation_id or 0,
                                 self.current_invocation_progress is not None, self.current_invocation_progress or 0.0,
                                 len(self.groups)))
        self.worker_resources.serialize(stream)
        _serialize_string(self.hwid, stream)
        _serialize_string(self.last_address, stream)
        for group in self.groups:
            _serialize_string(group, stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "WorkerData":
        worker_id, last_seen_timestamp, state_value, type_value, \
            has_current_invocation_node_id, current_invocation_node_id, \
            has_current_invocation_task_id, current_invocation_task_id, \
            has_current_invocation_id, current_invocation_id, \
            has_current_invocation_progress, current_invocation_progress, \
            groups_count = struct.unpack('>QQII?Q?Q?Q?dQ', stream.readexactly(68))
        if not has_current_invocation_node_id:
            current_invocation_node_id = None
        if not has_current_invocation_task_id:
            current_invocation_task_id = None
        if not has_current_invocation_id:
            current_invocation_id = None
        if not has_current_invocation_progress:
            current_invocation_progress = None
        worker_resources = WorkerResources.deserialize(stream)
        hwid = _deserialize_string(stream)
        last_address = _deserialize_string(stream)
        groups = set()
        for i in range(groups_count):
            groups.add(_deserialize_string(stream))
        return WorkerData(worker_id, worker_resources, hwid, last_address, last_seen_timestamp, WorkerState(state_value), WorkerType(type_value),
                          current_invocation_node_id, current_invocation_task_id, current_invocation_id, current_invocation_progress, groups)


@dataclass
class WorkerBatchData(IBufferSerializable):
    db_uid: int
    workers: Dict[int, WorkerData]

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQ', self.db_uid, len(self.workers)))
        for task in self.workers.values():
            task.serialize(stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "WorkerBatchData":
        db_uid, tasks_count, = struct.unpack('>QQ', stream.readexactly(16))
        workers = {}
        for i in range(tasks_count):
            task_data = WorkerData.deserialize(stream)
            workers[task_data.id] = task_data
        return WorkerBatchData(db_uid, workers)


@dataclass
class TaskGroupStatisticsData(IBufferSerializable):
    tasks_done: int
    tasks_in_progress: int
    tasks_with_error: int
    tasks_total: int

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQQQ', self.tasks_done or 0, self.tasks_in_progress or 0, self.tasks_with_error or 0, self.tasks_total or 0))

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "TaskGroupStatisticsData":
        tasks_done, tasks_in_progress, tasks_with_error, tasks_total = struct.unpack('>QQQQ', stream.readexactly(32))
        return TaskGroupStatisticsData(tasks_done, tasks_in_progress, tasks_with_error, tasks_total)


@dataclass
class TaskGroupData(IBufferSerializable):
    name: str
    creation_timestamp: int
    state: TaskGroupArchivedState
    priority: float
    statistics: Optional[TaskGroupStatisticsData]

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QId?', self.creation_timestamp, self.state.value, self.priority, self.statistics is not None))
        _serialize_string(self.name, stream)
        if self.statistics is not None:
            self.statistics.serialize(stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "TaskGroupData":
        ctimestamp, state_value, priority, has_statistics = struct.unpack('>QId?', stream.readexactly(21))
        name = _deserialize_string(stream)
        if has_statistics:
            statistics = TaskGroupStatisticsData.deserialize(stream)
        else:
            statistics = None
        return TaskGroupData(name, ctimestamp, TaskGroupArchivedState(state_value), priority, statistics)


@dataclass
class TaskGroupBatchData(IBufferSerializable):
    db_uid: int
    task_groups: Dict[str, TaskGroupData]

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQ', self.db_uid, len(self.task_groups)))
        for task_group in self.task_groups.values():
            task_group.serialize(stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "TaskGroupBatchData":
        db_uid, tasks_count, = struct.unpack('>QQ', stream.readexactly(16))
        task_groups = {}
        for i in range(tasks_count):
            group_data = TaskGroupData.deserialize(stream)
            task_groups[group_data.name] = group_data
        return TaskGroupBatchData(db_uid, task_groups)


@dataclass
class UiData(IBufferSerializable):  # Deprecated, should not be used any more
    db_uid: int
    graph_data: Optional[NodeGraphStructureData]
    tasks: Optional[TaskBatchData]
    workers: Optional[WorkerBatchData]
    task_groups: Optional[TaskGroupBatchData]

    def serialize(self, stream):
        buffer = BytesIO()

        buffer.write(struct.pack('>Q', self.db_uid))
        for data in (self.graph_data, self.tasks, self.workers, self.task_groups):
            buffer.write(struct.pack('>?', data is not None))
            if data is not None:
                data.serialize(buffer)

        lzdata = lz4.frame.compress(buffer.getbuffer())
        stream.write(struct.pack('>Q', len(lzdata)))
        stream.write(lzdata)

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "UiData":
        buffer = BytesIO(lz4.frame.decompress(stream.readexactly(struct.unpack('>Q', stream.readexactly(8))[0])))
        reader = BufferedReader(io.BufferedReader(buffer), 8192)

        db_uid, = struct.unpack('>Q', reader.readexactly(8))
        datas = []
        for data_type in (NodeGraphStructureData, TaskBatchData, WorkerBatchData, TaskGroupBatchData):
            if struct.unpack('>?', reader.readexactly(1))[0]:
                datas.append(data_type.deserialize(reader))
            else:
                datas.append(None)

        assert len(datas), 4
        return UiData(db_uid, datas[0], datas[1], datas[2], datas[3])

    def __repr__(self):
        return f'{self.graph_data} :::: {self.tasks}'  # TODO: this is not a good representation at all


@dataclass
class IncompleteInvocationLogData(IBufferSerializable):
    invocation_id: int
    worker_id: int
    invocation_runtime: Optional[float]

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQ?d', self.invocation_id, self.worker_id, self.invocation_runtime is not None, self.invocation_runtime or 0.0))

    @classmethod
    def deserialize(cls, stream: BufferedReader):
        i_id, w_id, has_i_rt, i_rt = struct.unpack('>QQ?d', stream.readexactly(25))
        if not has_i_rt:
            i_rt = None
        return IncompleteInvocationLogData(i_id, w_id, i_rt)


@dataclass
class InvocationLogData(IBufferSerializable):
    invocation_id: int
    worker_id: int
    invocation_runtime: Optional[float]
    task_id: int
    node_id: int
    invocation_state: InvocationState
    return_code: Optional[int]
    stdout: str
    stderr: str

    def serialize(self, stream: BufferedIOBase):
        stream.write(struct.pack('>QQ?dQQI?Q', self.invocation_id, self.worker_id, self.invocation_runtime is not None,
                                 self.invocation_runtime or 0.0, self.task_id, self.node_id, self.invocation_state.value,
                                 self.return_code is not None, self.return_code or 0))
        _serialize_string(self.stdout, stream)
        _serialize_string(self.stderr, stream)

    @classmethod
    def deserialize(cls, stream: BufferedReader):
        i_id, w_id, has_i_rt, i_rt, t_id, n_id, i_s_raw, has_ret_code, ret_code = struct.unpack('>QQ?dQQI?Q', stream.readexactly(54))
        if not has_i_rt:
            i_rt = None
        if not has_ret_code:
            ret_code = None
        stdout = _deserialize_string(stream)
        stderr = _deserialize_string(stream)
        return InvocationLogData(i_id, w_id, i_rt, t_id, n_id, InvocationState(i_s_raw), ret_code, stdout, stderr)



import os
import shlex
import re
from math import sqrt, floor, ceil

from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.uidata import NodeParameterType
from lifeblood.invocationjob import InvocationJob
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment

from typing import Iterable


def node_class():
    return Ffmpeg


description = \
'''uses imagemagick on a set of images
currently can only do montage
Args:
    sequence: list attribute to process as image source
              can either be a list of strings - one montage will be generated
              or a list of lists of strings - a montage will be generated 
                  for each bunch of images from inner lists with the same index:
                  [[a1, a2, a3], [b1,b2,b3]] will result in montages: [a1b1, a2b2, a3b3]
                                               
    hint tile count: generally tile count is guessed, but it can be hinted as well
    background: set background color, or none
    force depth: force depth of output image (if supported by format)
    force colorspace: force colorspace of output image (if supported by format)
    custom arguments: custom arguments to provide to montage command (in the end)
    use custom imagemagic path: by default it's assumed that imagemagick is available in PATH
                                you can provide specific bin directory where to look for montage command
    movie path: where to save output picture.
                if "replace # with padded frame" is set - ### will be replaced with properly padded frame number
                note that "frames" attribute must have enough values to provide frame values
                for all montages provided by "sequence"
                meaning number of frames must be >= then minimal sublist of "sequence"  
'''


class Ffmpeg(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'imagemagick'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'image', 'magic', 'stock', 'convert', 'montage'

    @classmethod
    def type_name(cls) -> str:
        return 'imagemagick'

    @classmethod
    def description(cls) -> str:
        return description

    def __init__(self, name):
        super(Ffmpeg, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.129, 0.239, 0.569)
            mode_param = ui.add_parameter('mode', 'mode', NodeParameterType.STRING, 'montage').add_menu((('montage', 'montage'), ('convert', 'convert')))

            ui.add_parameter('sequence', 'sequence attribute name', NodeParameterType.STRING, 'sequence')

            # montage parameters
            ui.add_parameter('bordersize', 'border size (pixels)', NodeParameterType.INT, 0)\
                .set_value_limits(value_min=0)\
                .append_visibility_condition(mode_param, '==', 'montage')
            with ui.parameters_on_same_line_block():
                dohint = ui.add_parameter('use tile hint', 'hint tile count', NodeParameterType.BOOL, False)
                dohint.append_visibility_condition(mode_param, '==', 'montage')
                ui.add_parameter('tile hint', 'hint', NodeParameterType.INT, 3)\
                    .append_visibility_condition(dohint, '==', True)\
                    .append_visibility_condition(mode_param, '==', 'montage')\
                    .set_value_limits(value_min=1)
                ui.add_parameter('tile hint is', '', NodeParameterType.STRING, 'col')\
                    .add_menu((('columns', 'col'), ('rows', 'row')))\
                    .append_visibility_condition(dohint, '==', True) \
                    .append_visibility_condition(mode_param, '==', 'montage')

            # convert parameters
            # are there any convert-specific parameters?

            # common parameters
            with ui.parameters_on_same_line_block():
                bgp = ui.add_parameter('background', 'background', NodeParameterType.BOOL, False)
                ui.add_parameter('background colorr', '', NodeParameterType.FLOAT, 0.0).append_visibility_condition(bgp, '==', True).set_value_limits(0.0, 1.0)
                ui.add_parameter('background colorg', '', NodeParameterType.FLOAT, 0.0).append_visibility_condition(bgp, '==', True).set_value_limits(0.0, 1.0)
                ui.add_parameter('background colorb', '', NodeParameterType.FLOAT, 0.0).append_visibility_condition(bgp, '==', True).set_value_limits(0.0, 1.0)
                ui.add_parameter('background colora', '', NodeParameterType.FLOAT, 1.0).append_visibility_condition(bgp, '==', True).set_value_limits(0.0, 1.0)
            with ui.parameters_on_same_line_block():
                dptp = ui.add_parameter('dodepth', 'force depth', NodeParameterType.BOOL, False)
                ui.add_parameter('depth', '', NodeParameterType.INT, 8)\
                    .add_menu((('8bit', 8), ('16bit', 16), ('32bit', 32)))\
                    .append_visibility_condition(dptp, '==', True)
            with ui.parameters_on_same_line_block():
                cspp = ui.add_parameter('docolorspace', 'force colorspace', NodeParameterType.BOOL, False)
                ui.add_parameter('colorspace', '', NodeParameterType.STRING, 'sRGB')\
                    .add_menu((('CMY', 'CMY'), ('CMYK', 'CMYK'), ('Gray', 'Gray'),
                               ('HCL', 'HCL'), ('HCLp', 'HCLp'), ('HSB', 'HSB'),
                               ('HSI', 'HSI'), ('HSL', 'HSL'), ('HSV', 'HSV'),
                               ('HWB', 'HWB'), ('Jzazbz', 'Jzazbz'), ('Lab', 'Lab'),
                               ('LCHab', 'LCHab'), ('LCHuv', 'LCHuv'), ('LMS', 'LMS'),
                               ('Log', 'Log'), ('Luv', 'Luv'), ('OHTA', 'OHTA'),
                               ('Rec601YCbCr', 'Rec601YCbCr'), ('Rec709YCbCr', 'Rec709YCbCr'), ('RGB', 'RGB'),
                               ('scRGB', 'scRGB'), ('sRGB', 'sRGB'), ('Transparent', 'Transparent'),
                               ('xyY', 'xyY'), ('XYZ', 'XYZ'), ('YCbCr', 'YCbCr'),
                               ('YCC', 'YCC'), ('YDbDr', 'YDbDr'), ('YIQ', 'YIQ'),
                               ('YPbPr', 'YPbPr'), ('YUV', 'YUV')
                               ))\
                    .append_visibility_condition(cspp, '==', True)
            ui.add_parameter('custom args', 'custom arguments', NodeParameterType.STRING, '')
            with ui.parameters_on_same_line_block():
                checkbox = ui.add_parameter('use custom path', 'use custom imagemagic path', NodeParameterType.BOOL, False)
                ui.add_parameter('imagemagic bin dir', 'path to bin dir', NodeParameterType.STRING, '').append_visibility_condition(checkbox, '==', True)
            with ui.parameters_on_same_line_block():
                ui.add_parameter('outpath', 'movie path', NodeParameterType.STRING, '/tmp/output.####.png')
                ui.add_parameter('outpath has #', 'replace # with padded frame', NodeParameterType.BOOL, True)

    def process_task(self, context) -> ProcessingResult:
        attributes = context.task_attributes()
        mode = context.param_value('mode')
        sequence_attrib_name = context.param_value('sequence')
        if mode == 'montage':
            if sequence_attrib_name not in attributes:
                raise ProcessingError(f'required attribute "{sequence_attrib_name}" not found')
            sequence = attributes[sequence_attrib_name]
            result_to_file = False
            if isinstance(sequence[0], str):  # we have sequence of frames -
                # treat it as parts of same mosaic in case of mosaic
                sequence = [[x] for x in sequence]  # just pack elements into lists and adapt for general case
                result_to_file = True
        elif mode == 'convert':
            # logic is - if sequence attrib exists - take it, if no - use 'file'
            if sequence_attrib_name in attributes:
                sequence = [attributes[sequence_attrib_name]]  # 2 nested lists to unify processing with montage mode
                result_to_file = False
            elif 'file' in attributes:
                sequence = [[attributes['file']]]  # 2 nested lists to unify processing with montage mode
                result_to_file = True
            else:
                raise ProcessingError(f'either "file" or "{sequence_attrib_name}" attribute must exist')
        frames = []
        if 'frames' in attributes:
            frames = attributes['frames']

        assert isinstance(sequence, list), 'sequence attribute is supposed to be list of frames, or list of sequences'

        outpath = context.param_value('outpath')
        replace_hashes = context.param_value('outpath has #')

        if context.param_value('use custom path'):
            bin = os.path.join(context.param_value('imagemagic bin dir'), mode)
        else:
            bin = mode

        custom_args = shlex.split(context.param_value('custom args'))

        hint = None
        border = None
        if mode == 'montage':
            if context.param_value('use tile hint'):
                if context.param_value('tile hint is') == 'col':
                    hint = f'{context.param_value("tile hint")}x'
                else:  # row
                    hint = f'x{context.param_value("tile hint")}'
            border = context.param_value('bordersize')

        assert isinstance(sequence[0], list)

        # construct a mosaic from one image from each subsequences
        numframes = min(len(x) for x in sequence)
        if replace_hashes and len(frames) < numframes:
            raise ProcessingError(f'frame list must have at least {numframes} frames to properly resolve output name')

        # just in case let's keep this script as py2-3 compatible as possible. why? actually there's no reason, unless it can be run from within some specific DCC, which it should
        script = f'import os\n' \
                 f'import re\n' \
                 f'import subprocess\n' \
                 f'sequence = {repr(sequence)}\n' \
                 f'frames = {repr(frames)}\n' \
                 f'outpath = {repr(outpath)}\n' \
                 f'if not os.path.exists(os.path.dirname(outpath)):\n' \
                 f'    try:\n' \
                 f'        os.makedirs(os.path.dirname(outpath))\n' \
                 f'    except OSError as e:\n' \
                 f'        import errno\n' \
                 f'        if e.errno != errno.EEXIST:\n' \
                 f'            raise\n' \
                 f'for i in range({numframes}):\n' \
                 f'    print("ALF_PROGRESS {{}}%".format(int(i*100.0/{numframes})))\n' \
                 f'    args = [{repr(bin)}]\n' \
                 f'    args += [x[i] for x in sequence]\n'
        if hint is not None:
            script += f"    args += ['-tile', {repr(hint)}]\n"
        if border is not None:
            script += f"    args += ['-geometry', f'+{border}+{border}']\n"
        if not context.param_value('background'):
            script += f'    args += ["-background", "none"]\n'
        else:
            r = context.param_value('background colorr') * 100
            g = context.param_value('background colorg') * 100
            b = context.param_value('background colorb') * 100
            a = context.param_value('background colora')
            script += f'    args += ["-background", "rgba({r}%,{g}%,{b}%,{a})"]\n'
            if mode == 'convert':
                script += f'    args += ["-flatten"]\n'
        if context.param_value('dodepth'):
            script += f'    args += ["-depth", "{context.param_value("depth")}"]\n'
        if context.param_value('docolorspace'):
            script += f'    args += ["-colorspace", {repr(context.param_value("colorspace"))}]\n'
        script += f"    args += {repr(custom_args)}\n"
        if replace_hashes:
            script += "    args.append(re.sub('#+', lambda m: '{{:0{}d}}'.format(len(m.group(0))).format(frames[i]), outpath))\n"
        else:
            script += "    args.append(outpath)\n"
        script += f'    subprocess.Popen(args).wait()\n'

        job = InvocationJob(['python', ':/work_to_do.py'])
        job.set_extra_file('work_to_do.py', script)
        res = ProcessingResult(job)
        if result_to_file:  # saving to "file" attribute a single path
            if replace_hashes:
                res.set_attribute('file', re.sub('#+', lambda m: f'{frames[0]:0{len(m.group(0))}d}', outpath))
            else:
                res.set_attribute('file', outpath)
        else:  # otherwise save a sequence of files into sequence attribute
            if replace_hashes:
                res.set_attribute('sequence', [re.sub('#+', lambda m: f'{i:0{len(m.group(0))}d}', outpath) for i in frames])
            else:
                res.set_attribute('sequence', [outpath for i in frames])  # this is a weird case...
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



from typing import Tuple


class ConfigurationError(RuntimeError):
    pass


class SchedulerConfigurationError(ConfigurationError):
    pass


class NodeNotReadyToProcess(Exception):
    pass


class AlreadyRunning(RuntimeError):
    pass


class NotEnoughResources(RuntimeError):
    pass


class WorkerNotAvailable(RuntimeError):
    pass


class ProcessInitializationError(RuntimeError):
    pass


class IncompleteReadError(ConnectionError):
    pass


class NotSubscribedError(RuntimeError):
    """
    UI State Accessor related exception: signifies that there is no subscription for the requested group
    """
    pass


class CouldNotNegotiateProtocolVersion(RuntimeError):
    def __init__(self,
                 our_supported_versions: Tuple[Tuple[int, int], ...],
                 their_supported_versions: Tuple[Tuple[int, int], ...]):
        self.__ours = our_supported_versions
        self.__theirs = their_supported_versions

    def __repr__(self):
        return f'<{self.__class__.__name__}: ours:{self.__ours}, theirs:{self.__theirs}>'


class InvocationMessageError(RuntimeError):
    pass


class InvocationMessageWrongInvocationId(InvocationMessageError):
    pass


class InvocationMessageAddresseeTimeout(InvocationMessageError):
    pass
